from esphome import automation
from esphome.automation import maybe_simple_id
import esphome.codegen as cg
from esphome.components import mqtt, web_server
import esphome.config_validation as cv
from esphome.const import (
    CONF_CODE,
    CONF_ENTITY_CATEGORY,
    CONF_ICON,
    CONF_ID,
    CONF_MQTT_ID,
    CONF_ON_STATE,
    CONF_WEB_SERVER,
)
from esphome.core import CORE, CoroPriority, coroutine_with_priority
from esphome.core.entity_helpers import (
    entity_duplicate_validator,
    queue_entity_register,
    setup_entity,
)
from esphome.cpp_generator import MockObjClass

CODEOWNERS = ["@grahambrown11", "@hwstar"]
IS_PLATFORM_COMPONENT = True

CONF_ON_TRIGGERED = "on_triggered"
CONF_ON_CLEARED = "on_cleared"
CONF_ON_ARMING = "on_arming"
CONF_ON_PENDING = "on_pending"
CONF_ON_ARMED_HOME = "on_armed_home"
CONF_ON_ARMED_NIGHT = "on_armed_night"
CONF_ON_ARMED_AWAY = "on_armed_away"
CONF_ON_DISARMED = "on_disarmed"
CONF_ON_CHIME = "on_chime"
CONF_ON_READY = "on_ready"

alarm_control_panel_ns = cg.esphome_ns.namespace("alarm_control_panel")
AlarmControlPanel = alarm_control_panel_ns.class_("AlarmControlPanel", cg.EntityBase)

StateAnyForwarder = alarm_control_panel_ns.class_("StateAnyForwarder")
StateEnterForwarder = alarm_control_panel_ns.class_("StateEnterForwarder")
AlarmControlPanelState = alarm_control_panel_ns.enum("AlarmControlPanelState")

ArmAwayAction = alarm_control_panel_ns.class_("ArmAwayAction", automation.Action)
ArmHomeAction = alarm_control_panel_ns.class_("ArmHomeAction", automation.Action)
ArmNightAction = alarm_control_panel_ns.class_("ArmNightAction", automation.Action)
DisarmAction = alarm_control_panel_ns.class_("DisarmAction", automation.Action)
PendingAction = alarm_control_panel_ns.class_("PendingAction", automation.Action)
TriggeredAction = alarm_control_panel_ns.class_("TriggeredAction", automation.Action)
ChimeAction = alarm_control_panel_ns.class_("ChimeAction", automation.Action)
ReadyAction = alarm_control_panel_ns.class_("ReadyAction", automation.Action)

AlarmControlPanelCondition = alarm_control_panel_ns.class_(
    "AlarmControlPanelCondition", automation.Condition
)

_ALARM_CONTROL_PANEL_SCHEMA = (
    cv.ENTITY_BASE_SCHEMA.extend(web_server.WEBSERVER_SORTING_SCHEMA)
    .extend(cv.MQTT_COMMAND_COMPONENT_SCHEMA)
    .extend(
        {
            cv.OnlyWith(CONF_MQTT_ID, "mqtt"): cv.declare_id(
                mqtt.MQTTAlarmControlPanelComponent
            ),
            cv.Optional(CONF_ON_STATE): automation.validate_automation({}),
            cv.Optional(CONF_ON_TRIGGERED): automation.validate_automation({}),
            cv.Optional(CONF_ON_ARMING): automation.validate_automation({}),
            cv.Optional(CONF_ON_PENDING): automation.validate_automation({}),
            cv.Optional(CONF_ON_ARMED_HOME): automation.validate_automation({}),
            cv.Optional(CONF_ON_ARMED_NIGHT): automation.validate_automation({}),
            cv.Optional(CONF_ON_ARMED_AWAY): automation.validate_automation({}),
            cv.Optional(CONF_ON_DISARMED): automation.validate_automation({}),
            cv.Optional(CONF_ON_CLEARED): automation.validate_automation({}),
            cv.Optional(CONF_ON_CHIME): automation.validate_automation({}),
            cv.Optional(CONF_ON_READY): automation.validate_automation({}),
        }
    )
)


_ALARM_CONTROL_PANEL_SCHEMA.add_extra(entity_duplicate_validator("alarm_control_panel"))


def alarm_control_panel_schema(
    class_: MockObjClass,
    *,
    entity_category: str = cv.UNDEFINED,
    icon: str = cv.UNDEFINED,
) -> cv.Schema:
    schema = {
        cv.GenerateID(): cv.declare_id(class_),
    }

    for key, default, validator in [
        (CONF_ENTITY_CATEGORY, entity_category, cv.entity_category),
        (CONF_ICON, icon, cv.icon),
    ]:
        if default is not cv.UNDEFINED:
            schema[cv.Optional(key, default=default)] = validator

    return _ALARM_CONTROL_PANEL_SCHEMA.extend(schema)


ALARM_CONTROL_PANEL_ACTION_SCHEMA = maybe_simple_id(
    {
        cv.GenerateID(): cv.use_id(AlarmControlPanel),
        cv.Optional(CONF_CODE): cv.templatable(cv.string),
    }
)

ALARM_CONTROL_PANEL_CONDITION_SCHEMA = maybe_simple_id(
    {
        cv.GenerateID(): cv.use_id(AlarmControlPanel),
    }
)


_CALLBACK_AUTOMATIONS = (
    automation.CallbackAutomation(
        CONF_ON_STATE, "add_on_state_callback", forwarder=StateAnyForwarder
    ),
    automation.CallbackAutomation(
        CONF_ON_TRIGGERED,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(
            AlarmControlPanelState.ACP_STATE_TRIGGERED
        ),
    ),
    automation.CallbackAutomation(
        CONF_ON_ARMING,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(AlarmControlPanelState.ACP_STATE_ARMING),
    ),
    automation.CallbackAutomation(
        CONF_ON_PENDING,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(
            AlarmControlPanelState.ACP_STATE_PENDING
        ),
    ),
    automation.CallbackAutomation(
        CONF_ON_ARMED_HOME,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(
            AlarmControlPanelState.ACP_STATE_ARMED_HOME
        ),
    ),
    automation.CallbackAutomation(
        CONF_ON_ARMED_NIGHT,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(
            AlarmControlPanelState.ACP_STATE_ARMED_NIGHT
        ),
    ),
    automation.CallbackAutomation(
        CONF_ON_ARMED_AWAY,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(
            AlarmControlPanelState.ACP_STATE_ARMED_AWAY
        ),
    ),
    automation.CallbackAutomation(
        CONF_ON_DISARMED,
        "add_on_state_callback",
        forwarder=StateEnterForwarder.template(
            AlarmControlPanelState.ACP_STATE_DISARMED
        ),
    ),
    automation.CallbackAutomation(CONF_ON_CLEARED, "add_on_cleared_callback"),
    automation.CallbackAutomation(CONF_ON_CHIME, "add_on_chime_callback"),
    automation.CallbackAutomation(CONF_ON_READY, "add_on_ready_callback"),
)


@setup_entity("alarm_control_panel")
async def setup_alarm_control_panel_core_(var, config):
    await automation.build_callback_automations(var, config, _CALLBACK_AUTOMATIONS)
    if web_server_config := config.get(CONF_WEB_SERVER):
        await web_server.add_entity_config(var, web_server_config)
    if mqtt_id := config.get(CONF_MQTT_ID):
        mqtt_ = cg.new_Pvariable(mqtt_id, var)
        await mqtt.register_mqtt_component(mqtt_, config)


async def register_alarm_control_panel(var, config):
    if not CORE.has_id(config[CONF_ID]):
        var = cg.Pvariable(config[CONF_ID], var)
    queue_entity_register("alarm_control_panel", config)
    CORE.register_platform_component("alarm_control_panel", var)
    await setup_alarm_control_panel_core_(var, config)


async def new_alarm_control_panel(config, *args):
    var = cg.new_Pvariable(config[CONF_ID], *args)
    await register_alarm_control_panel(var, config)
    return var


@automation.register_action(
    "alarm_control_panel.arm_away",
    ArmAwayAction,
    ALARM_CONTROL_PANEL_ACTION_SCHEMA,
    synchronous=True,
)
async def alarm_action_arm_away_to_code(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    var = cg.new_Pvariable(action_id, template_arg, paren)
    if code_config := config.get(CONF_CODE):
        templatable_ = await cg.templatable(code_config, args, cg.std_string)
        cg.add(var.set_code(templatable_))
    return var


@automation.register_action(
    "alarm_control_panel.arm_home",
    ArmHomeAction,
    ALARM_CONTROL_PANEL_ACTION_SCHEMA,
    synchronous=True,
)
async def alarm_action_arm_home_to_code(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    var = cg.new_Pvariable(action_id, template_arg, paren)
    if code_config := config.get(CONF_CODE):
        templatable_ = await cg.templatable(code_config, args, cg.std_string)
        cg.add(var.set_code(templatable_))
    return var


@automation.register_action(
    "alarm_control_panel.arm_night",
    ArmNightAction,
    ALARM_CONTROL_PANEL_ACTION_SCHEMA,
    synchronous=True,
)
async def alarm_action_arm_night_to_code(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    var = cg.new_Pvariable(action_id, template_arg, paren)
    if CONF_CODE in config:
        templatable_ = await cg.templatable(config[CONF_CODE], args, cg.std_string)
        cg.add(var.set_code(templatable_))
    return var


@automation.register_action(
    "alarm_control_panel.disarm",
    DisarmAction,
    ALARM_CONTROL_PANEL_ACTION_SCHEMA,
    synchronous=True,
)
async def alarm_action_disarm_to_code(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    var = cg.new_Pvariable(action_id, template_arg, paren)
    if code_config := config.get(CONF_CODE):
        templatable_ = await cg.templatable(code_config, args, cg.std_string)
        cg.add(var.set_code(templatable_))
    return var


@automation.register_action(
    "alarm_control_panel.pending",
    PendingAction,
    ALARM_CONTROL_PANEL_ACTION_SCHEMA,
    synchronous=True,
)
async def alarm_action_pending_to_code(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    return cg.new_Pvariable(action_id, template_arg, paren)


@automation.register_action(
    "alarm_control_panel.triggered",
    TriggeredAction,
    ALARM_CONTROL_PANEL_ACTION_SCHEMA,
    synchronous=True,
)
async def alarm_action_trigger_to_code(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    return cg.new_Pvariable(action_id, template_arg, paren)


@automation.register_action(
    "alarm_control_panel.chime",
    ChimeAction,
    ALARM_CONTROL_PANEL_ACTION_SCHEMA,
    synchronous=True,
)
async def alarm_action_chime_to_code(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    return cg.new_Pvariable(action_id, template_arg, paren)


@automation.register_action(
    "alarm_control_panel.ready",
    ReadyAction,
    ALARM_CONTROL_PANEL_ACTION_SCHEMA,
    synchronous=True,
)
@automation.register_condition(
    "alarm_control_panel.ready",
    AlarmControlPanelCondition,
    ALARM_CONTROL_PANEL_CONDITION_SCHEMA,
)
async def alarm_action_ready_to_code(config, action_id, template_arg, args):
    paren = await cg.get_variable(config[CONF_ID])
    return cg.new_Pvariable(action_id, template_arg, paren)


@automation.register_condition(
    "alarm_control_panel.is_armed",
    AlarmControlPanelCondition,
    ALARM_CONTROL_PANEL_CONDITION_SCHEMA,
)
async def alarm_control_panel_is_armed_to_code(
    config, condition_id, template_arg, args
):
    paren = await cg.get_variable(config[CONF_ID])
    return cg.new_Pvariable(condition_id, template_arg, paren)


@coroutine_with_priority(CoroPriority.CORE)
async def to_code(config):
    cg.add_global(alarm_control_panel_ns.using)



from collections.abc import Callable
from contextlib import AbstractContextManager
from dataclasses import dataclass
import importlib
import importlib.abc
import importlib.resources
import importlib.util
import logging
from pathlib import Path
import sys
from types import ModuleType
from typing import TYPE_CHECKING, Any

from esphome.const import SOURCE_FILE_EXTENSIONS
from esphome.types import ConfigType

if TYPE_CHECKING:
    from esphome.cpp_generator import MockObjClass

# `esphome.core.config` is imported lazily in `_lookup_module` when the
# "esphome" pseudo-component is first resolved. It pulls in
# `esphome.automation` and `esphome.config_validation`, which together
# dominate `esphome.__main__` startup cost when loaded eagerly.
# `esphome.cpp_generator` is similarly avoided at module scope; it pulls
# in `esphome.yaml_util` and is only needed for the `MockObjClass` type
# annotation, which is resolved lazily via `TYPE_CHECKING`.

_LOGGER = logging.getLogger(__name__)


@dataclass(frozen=True, order=True)
class FileResource:
    package: str
    resource: str

    def path(self) -> AbstractContextManager[Path]:
        return importlib.resources.as_file(
            importlib.resources.files(self.package) / self.resource
        )


class ComponentManifest:
    def __init__(self, module: ModuleType, recursive_sources: bool = False):
        self.module = module
        self.recursive_sources = recursive_sources

    @property
    def package(self) -> str:
        """Return the package name the module is contained in.

        Examples:
        - esphome/components/gpio/__init__.py -> esphome.components.gpio
        - esphome/components/gpio/switch/__init__.py -> esphome.components.gpio.switch
        - esphome/components/a4988/stepper.py -> esphome.components.a4988
        """
        return self.module.__package__

    @property
    def is_platform(self) -> bool:
        return len(self.module.__name__.split(".")) == 4

    @property
    def is_platform_component(self) -> bool:
        return getattr(self.module, "IS_PLATFORM_COMPONENT", False)

    @property
    def is_target_platform(self) -> bool:
        return getattr(self.module, "IS_TARGET_PLATFORM", False)

    @property
    def config_schema(self) -> Any | None:
        return getattr(self.module, "CONFIG_SCHEMA", None)

    @property
    def multi_conf(self) -> bool:
        return getattr(self.module, "MULTI_CONF", False)

    @property
    def multi_conf_no_default(self) -> bool:
        return getattr(self.module, "MULTI_CONF_NO_DEFAULT", False)

    @property
    def to_code(self) -> Callable[[Any], None] | None:
        return getattr(self.module, "to_code", None)

    @property
    def dependencies(self) -> list[str]:
        return getattr(self.module, "DEPENDENCIES", [])

    @property
    def conflicts_with(self) -> list[str]:
        return getattr(self.module, "CONFLICTS_WITH", [])

    @property
    def auto_load(
        self,
    ) -> list[str] | Callable[[], list[str]] | Callable[[ConfigType], list[str]]:
        return getattr(self.module, "AUTO_LOAD", [])

    @property
    def codeowners(self) -> list[str]:
        return getattr(self.module, "CODEOWNERS", [])

    @property
    def aliases(self) -> list[str]:
        """Legacy names that should transparently route to this component.

        See the :func:`_build_alias_map` documentation for how aliases are
        discovered (AST scan, no execution) and registered both for the YAML
        loader (top-level key rename in :mod:`esphome.config`) and for
        Python imports (``sys.meta_path`` finder, below).
        """
        return getattr(self.module, "ALIASES", [])

    @property
    def alias_removal_version(self) -> str | None:
        """Optional ESPHome version when the alias warning becomes a hard error.

        Surfaced in the deprecation warning emitted by the YAML pre-pass so
        users know how long they have to migrate. ``None`` means the warning
        does not mention a specific version.
        """
        return getattr(self.module, "ALIAS_REMOVAL_VERSION", None)

    @property
    def instance_type(self) -> "MockObjClass | None":
        return getattr(self.module, "INSTANCE_TYPE", None)

    @property
    def final_validate_schema(self) -> Callable[[ConfigType], None] | None:
        """Components can declare a `FINAL_VALIDATE_SCHEMA` cv.Schema that gets called
        after the main validation. In that function checks across components can be made.

        Note that the function can't mutate the configuration - no changes are saved
        """
        return getattr(self.module, "FINAL_VALIDATE_SCHEMA", None)

    @property
    def legacy_config_migrate(self) -> Callable[[ConfigType], ConfigType | None] | None:
        """Optional `LEGACY_CONFIG_MIGRATE` callable on a platform component module.

        Called once, before platform entries are processed, with the raw top-level
        config for this domain. It may transform a pre-platform-format config (e.g.
        a bare list or legacy dict form) into the normalized list of `platform:`
        tagged entries and return it. Returning ``None`` means "already in the new
        format, leave untouched". This is an intentionally removable deprecation
        shim hook.
        """
        return getattr(self.module, "LEGACY_CONFIG_MIGRATE", None)

    @property
    def resources(self) -> list[FileResource]:
        """Return a list of all file resources defined in the package of this component.

        By default only files directly in the package directory are returned. Manifests
        constructed with ``recursive_sources=True`` also descend into non-subpackage
        subdirectories (subdirectories without an ``__init__.py``), so core code can
        live under ``esphome/core/<group>/`` without every component paying the cost.
        """
        ret: list[FileResource] = []

        # Get filter function for source files
        filter_source_files_func = getattr(self.module, "FILTER_SOURCE_FILES", None)

        # Get list of files to exclude
        excluded_files = (
            set(filter_source_files_func()) if filter_source_files_func else set()
        )

        root = importlib.resources.files(self.package)

        for child in root.iterdir():
            name = child.name
            if child.is_file():
                if Path(name).suffix not in SOURCE_FILE_EXTENSIONS:
                    continue
                if name in excluded_files:
                    continue
                ret.append(FileResource(self.package, name))
            elif self.recursive_sources and child.is_dir() and name != "__pycache__":
                # Skip Python subpackages — they load as their own components.
                if child.joinpath("__init__.py").is_file():
                    continue
                for sub in child.iterdir():
                    if not sub.is_file():
                        continue
                    if Path(sub.name).suffix not in SOURCE_FILE_EXTENSIONS:
                        continue
                    resource = f"{name}/{sub.name}"
                    if resource in excluded_files:
                        continue
                    ret.append(FileResource(self.package, resource))

        return ret


class ComponentMetaFinder(importlib.abc.MetaPathFinder):
    def __init__(
        self, components_path: Path, allowed_components: list[str] | None = None
    ) -> None:
        self._allowed_components = allowed_components
        self._finders = []
        for hook in sys.path_hooks:
            try:
                finder = hook(str(components_path))
            except ImportError:
                continue
            self._finders.append(finder)

    def find_spec(self, fullname: str, path: list[str] | None, target=None):
        if not fullname.startswith("esphome.components."):
            return None
        parts = fullname.split(".")
        if len(parts) != 3:
            # only handle direct components, not platforms
            # platforms are handled automatically when parent is imported
            return None
        component = parts[2]
        if (
            self._allowed_components is not None
            and component not in self._allowed_components
        ):
            return None

        for finder in self._finders:
            spec = finder.find_spec(fullname, target=target)
            if spec is not None:
                return spec
        return None


def clear_component_meta_finders():
    sys.meta_path = [x for x in sys.meta_path if not isinstance(x, ComponentMetaFinder)]


def install_meta_finder(
    components_path: Path, allowed_components: list[str] | None = None
):
    sys.meta_path.insert(0, ComponentMetaFinder(components_path, allowed_components))


def _lookup_module(domain: str, exception: bool) -> ComponentManifest | None:
    if domain in _COMPONENT_CACHE:
        return _COMPONENT_CACHE[domain]

    if domain == "esphome":
        import esphome.core.config

        manif = ComponentManifest(esphome.core.config, recursive_sources=True)
        _COMPONENT_CACHE[domain] = manif
        return manif

    # If `domain` is the legacy name of a renamed component, redirect to the
    # canonical module so the rest of the loader (and every caller of
    # `get_component(legacy)`) transparently sees the new component.
    alias_map = _get_alias_map()
    if domain in alias_map:
        canonical = alias_map[domain]
        manif = _lookup_module(canonical, exception)
        if manif is not None:
            _COMPONENT_CACHE[domain] = manif
        return manif

    try:
        module = importlib.import_module(f"esphome.components.{domain}")
    except ImportError as e:
        if exception:
            raise
        if "No module named" in str(e):
            _LOGGER.info(
                "Unable to import component %s: %s", domain, str(e), exc_info=False
            )
        else:
            _LOGGER.exception("Unable to import component %s:", domain)
        return None
    except Exception:  # pylint: disable=broad-except
        if exception:
            raise
        _LOGGER.exception("Unable to load component %s:", domain)
        return None

    manif = ComponentManifest(module)
    _COMPONENT_CACHE[domain] = manif
    return manif


def get_component(domain: str, exception: bool = False) -> ComponentManifest | None:
    assert "." not in domain
    return _lookup_module(domain, exception)


def get_platform(domain: str, platform: str) -> ComponentManifest | None:
    full = f"{platform}.{domain}"
    return _lookup_module(full, False)


_COMPONENT_CACHE: dict[str, ComponentManifest] = {}
CORE_COMPONENTS_PATH = (Path(__file__).parent / "components").resolve()


def _replace_component_manifest(domain: str, manifest: ComponentManifest) -> None:
    """Replace the cached manifest for a component.

    This is an intentionally-supported hook for the C++ test infrastructure
    to install ``ComponentManifestOverride`` wrappers.  Normal application
    code should never call this.
    """
    _COMPONENT_CACHE[domain] = manifest


# ---------------------------------------------------------------------------
# Component aliases (renamed-platform back-compat)
# ---------------------------------------------------------------------------
#
# A component can declare ``ALIASES = ["legacy_name"]`` (and optionally
# ``ALIAS_REMOVAL_VERSION = "YYYY.M.0"``) in its ``__init__.py``. Two
# integrations are then wired up automatically:
#
#   1. **Python imports** — a ``sys.meta_path`` finder (``_AliasFinder``)
#      intercepts ``esphome.components.<legacy>``/``...<legacy>.<sub>``
#      imports and resolves them against the canonical component so external
#      custom components that still import from the old path keep working.
#
#   2. **YAML loader** — ``_lookup_module`` consults the alias map so
#      ``get_component("legacy")`` returns the canonical manifest. The
#      ``esphome.config`` pre-pass uses the same map to rewrite legacy
#      top-level keys in the user's config (with a deprecation warning) so
#      dependency checks, schema validation and codegen all see only the
#      canonical name.
#
# Both lookups are populated by ``_build_alias_map``, which **AST-parses**
# every component's ``__init__.py`` rather than importing it. That keeps the
# cost low: scanning ~400 components on disk takes ~5 ms instead of the
# multi-second cost of executing every component's import side-effects.


_ALIAS_MAP_CACHE: dict[str, str] | None = None
_ALIAS_META_CACHE: dict[str, "AliasMeta"] | None = None


@dataclass(frozen=True)
class AliasMeta:
    """Metadata for a single deprecated alias entry.

    Used by the YAML pre-pass in :mod:`esphome.config` to produce a
    deprecation warning citing the canonical name and (optionally) the
    removal version declared by the canonical component.
    """

    canonical: str
    removal_version: str | None


def _ensure_alias_caches() -> None:
    """Populate both alias caches from a single directory scan.

    ``_build_alias_map`` returns both maps together, so building them in one
    shot avoids scanning every component's ``__init__.py`` twice when a run
    needs both the canonical map (loader) and the metadata map (config
    pre-pass).
    """
    global _ALIAS_MAP_CACHE, _ALIAS_META_CACHE
    if _ALIAS_MAP_CACHE is None or _ALIAS_META_CACHE is None:
        _ALIAS_MAP_CACHE, _ALIAS_META_CACHE = _build_alias_map()


def _get_alias_map() -> dict[str, str]:
    """Return the legacy-name → canonical-name map, building it lazily."""
    _ensure_alias_caches()
    return _ALIAS_MAP_CACHE


def get_alias_metadata() -> dict[str, AliasMeta]:
    """Return the legacy-name → :class:`AliasMeta` map (cached).

    Used by the YAML pre-pass to format a per-alias deprecation warning.
    """
    _ensure_alias_caches()
    return _ALIAS_META_CACHE


def _build_alias_map() -> tuple[dict[str, str], dict[str, AliasMeta]]:
    """Scan every core component dir for ``ALIASES`` declarations.

    Uses :mod:`ast` to read each component's ``__init__.py`` without
    executing it — component import side-effects (logger setup,
    namespace registration, etc.) shouldn't run just because we're
    enumerating aliases.

    Raises if the same alias is claimed by two canonical components, since
    silently picking one would cause non-deterministic routing depending on
    directory-iteration order. Also raises if an alias shadows an existing
    component package: that would hijack a live component domain and, in the
    self-alias case (alias == canonical), send ``_lookup_module`` into
    infinite recursion redirecting a domain to itself.
    """
    import ast

    alias_to_canonical: dict[str, str] = {}
    alias_to_meta: dict[str, AliasMeta] = {}

    if not CORE_COMPONENTS_PATH.is_dir():
        return alias_to_canonical, alias_to_meta

    for child in sorted(CORE_COMPONENTS_PATH.iterdir()):
        if not child.is_dir():
            continue
        init = child / "__init__.py"
        if not init.is_file():
            continue
        aliases, removal_version = _read_aliases(init, ast)
        if not aliases:
            continue
        canonical = child.name
        for alias in aliases:
            if (CORE_COMPONENTS_PATH / alias / "__init__.py").is_file():
                from esphome.core import EsphomeError

                raise EsphomeError(
                    f"Component alias '{alias}' (declared by '{canonical}') "
                    "shadows an existing component package of the same name. "
                    "An alias may only name a component that no longer exists."
                )
            if alias in alias_to_canonical:
                from esphome.core import EsphomeError

                raise EsphomeError(
                    f"Component alias '{alias}' is declared by both "
                    f"'{alias_to_canonical[alias]}' and '{canonical}'. "
                    "Each alias must map to exactly one canonical component."
                )
            alias_to_canonical[alias] = canonical
            alias_to_meta[alias] = AliasMeta(
                canonical=canonical, removal_version=removal_version
            )
    return alias_to_canonical, alias_to_meta


def _read_aliases(
    init_path: Path, ast_module: ModuleType
) -> tuple[list[str], str | None]:
    """Extract ``ALIASES`` and ``ALIAS_REMOVAL_VERSION`` from a component
    ``__init__.py`` via AST parsing.

    Only handles the simple ``NAME = [str_literal, ...]`` / ``NAME = "..."``
    forms — anything more dynamic (function call, conditional, etc.) is
    silently ignored. Components should keep their alias declarations
    static so this scanner can see them.
    """
    try:
        source = init_path.read_text(encoding="utf-8")
    except OSError as err:
        _LOGGER.warning(
            "Could not read %s while scanning for component aliases: %s",
            init_path,
            err,
        )
        return [], None

    # Cheap substring pre-filter: almost no component declares ALIASES, and
    # parsing every component __init__.py with ast is comparatively expensive.
    # Skip the parse entirely unless the token appears in the file at all.
    if "ALIASES" not in source:
        return [], None

    try:
        tree = ast_module.parse(source)
    except SyntaxError as err:
        _LOGGER.warning(
            "Could not parse %s while scanning for component aliases: %s",
            init_path,
            err,
        )
        return [], None

    aliases: list[str] = []
    removal_version: str | None = None

    for node in tree.body:
        if not isinstance(node, ast_module.Assign):
            continue
        for target in node.targets:
            if not isinstance(target, ast_module.Name):
                continue
            if target.id == "ALIASES" and isinstance(node.value, ast_module.List):
                aliases.extend(
                    elt.value
                    for elt in node.value.elts
                    if isinstance(elt, ast_module.Constant)
                    and isinstance(elt.value, str)
                )
            elif (
                target.id == "ALIAS_REMOVAL_VERSION"
                and isinstance(node.value, ast_module.Constant)
                and isinstance(node.value.value, str)
            ):
                removal_version = node.value.value
    return aliases, removal_version


class _AliasFinder(importlib.abc.MetaPathFinder):
    """``sys.meta_path`` finder that resolves legacy-component imports.

    Routes ``esphome.components.<alias>[.<submod>]`` to the canonical
    component's module/submodule of the same name, so external code that
    still imports ``from esphome.components.rp2040 import boards`` keeps
    working without the canonical component having to maintain a shim
    package on disk.

    The finder caches the resolved module in ``sys.modules`` under the
    legacy name on first lookup, so subsequent imports hit the cache and
    skip this finder entirely.
    """

    _PREFIX = "esphome.components."

    def find_spec(self, fullname, path, target=None):  # noqa: ARG002
        if not fullname.startswith(self._PREFIX):
            return None
        # Anything matching the ``esphome.components.`` prefix splits into at
        # least three parts, so ``parts[2]`` (the domain) always exists.
        parts = fullname.split(".")
        domain = parts[2]
        alias_map = _get_alias_map()
        if domain not in alias_map:
            return None

        parts[2] = alias_map[domain]
        canonical_fullname = ".".join(parts)
        try:
            canonical_module = importlib.import_module(canonical_fullname)
        except ModuleNotFoundError as err:
            # Only treat a missing *canonical target* as "no alias to
            # resolve" (let the normal import machinery report it). If some
            # other module is missing, the canonical exists but failed to
            # import one of its own dependencies — surface that real error
            # rather than masking it as an unresolved alias.
            if err.name == canonical_fullname:
                return None
            raise
        # Do NOT pre-populate ``sys.modules[fullname]`` here. Python's
        # ``_find_spec`` (in importlib._bootstrap) has an optimization that
        # detects ``name in sys.modules`` after a finder returns and prefers
        # ``sys.modules[name].__spec__`` over the finder's spec — for an
        # alias, that's the canonical module's own SourceFileLoader spec,
        # which Python then *re-loads*, defeating the aliasing. Letting
        # ``_load_unlocked`` populate sys.modules itself (via our
        # ``_AliasLoader.create_module``) sidesteps that branch.
        return importlib.util.spec_from_loader(fullname, _AliasLoader(canonical_module))


class _AliasLoader(importlib.abc.Loader):
    """No-op loader that returns the already-resolved canonical module.

    :class:`_AliasFinder` populates ``sys.modules`` itself; this loader
    just satisfies the :mod:`importlib` protocol so Python doesn't try to
    re-execute the module.
    """

    def __init__(self, module: ModuleType) -> None:
        self._module = module

    def create_module(self, spec):  # noqa: ARG002
        return self._module

    def exec_module(self, module):  # noqa: ARG002
        # Nothing to execute — the canonical module is already initialized.
        return None


# Register once at module load. Idempotent: re-installing the finder on
# repeated imports (e.g. by tests that reload `esphome.loader`) is a no-op
# because we check for an existing instance first.
def _install_alias_finder() -> None:
    for entry in sys.meta_path:
        if isinstance(entry, _AliasFinder):
            return
    sys.meta_path.append(_AliasFinder())


_install_alias_finder()



import esphome.codegen as cg
from esphome.components import sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_DISTANCE,
    CONF_LIGHTNING_ENERGY,
    ICON_FLASH,
    ICON_SIGNAL_DISTANCE_VARIANT,
    STATE_CLASS_MEASUREMENT,
    UNIT_KILOMETER,
)

from . import AS3935, CONF_AS3935_ID

DEPENDENCIES = ["as3935"]

CONFIG_SCHEMA = cv.Schema(
    {
        cv.GenerateID(CONF_AS3935_ID): cv.use_id(AS3935),
        cv.Optional(CONF_DISTANCE): sensor.sensor_schema(
            unit_of_measurement=UNIT_KILOMETER,
            icon=ICON_SIGNAL_DISTANCE_VARIANT,
            accuracy_decimals=1,
            state_class=STATE_CLASS_MEASUREMENT,
        ),
        cv.Optional(CONF_LIGHTNING_ENERGY): sensor.sensor_schema(
            icon=ICON_FLASH,
            accuracy_decimals=1,
        ),
    }
)


async def to_code(config):
    hub = await cg.get_variable(config[CONF_AS3935_ID])

    if distance_config := config.get(CONF_DISTANCE):
        sens = await sensor.new_sensor(distance_config)
        cg.add(hub.set_distance_sensor(sens))

    if lightning_energy_config := config.get(CONF_LIGHTNING_ENERGY):
        sens = await sensor.new_sensor(lightning_energy_config)
        cg.add(hub.set_energy_sensor(sens))



import logging
import re

import esphome.codegen as cg
from esphome.components import esp32, uart
import esphome.config_validation as cv
from esphome.const import (
    CONF_ID,
    CONF_NAME,
    CONF_PATTERN,
    CONF_PRIORITY,
    CONF_RECEIVE_TIMEOUT,
)
from esphome.core import CORE

_LOGGER = logging.getLogger(__name__)

CODEOWNERS = ["@SimonFischer04", "@Tomer27cz", "@latonita", "@PolarGoose"]
DEPENDENCIES = ["uart"]

CONF_DLMS_METER_ID = "dlms_meter_id"
CONF_DECRYPTION_KEY = "decryption_key"
CONF_AUTH_KEY = "auth_key"
CONF_OBIS_CODE = "obis_code"
CONF_CUSTOM_PATTERNS = "custom_patterns"
CONF_SKIP_CRC = "skip_crc"
CONF_DEFAULT_OBIS = "default_obis"
CONF_PROVIDER = "provider"

dlms_meter_component_ns = cg.esphome_ns.namespace("dlms_meter")
DlmsMeterComponent = dlms_meter_component_ns.class_(
    "DlmsMeterComponent", cg.Component, uart.UARTDevice
)


def obis_code(value):
    # Normalize the OBIS code to the strict A.B.C.D.E.F format
    bytes_list = parse_obis_code_bytes(value)
    return ".".join(str(b) for b in bytes_list)


def parse_obis_code_bytes(value):
    value = cv.string(value)
    normalized = re.sub(r"[\-\:\*]", ".", value)
    parts = normalized.split(".")
    if len(parts) < 5 or len(parts) > 6:
        raise cv.Invalid("OBIS code must have 5 or 6 parts")
    try:
        bytes_list = [int(p) for p in parts]
    except ValueError as exc:
        raise cv.Invalid("OBIS code parts must be integers") from exc
    for b in bytes_list:
        if b < 0 or b > 255:
            raise cv.Invalid("OBIS code parts must be between 0 and 255")
    if len(bytes_list) == 5:
        bytes_list.append(255)
    return bytes_list


def custom_pattern_dict(value):
    if isinstance(value, str):
        return {CONF_PATTERN: value}
    return value


def validate_custom_pattern(value):
    if CONF_DEFAULT_OBIS in value and CONF_NAME not in value:
        raise cv.Invalid(f"'{CONF_DEFAULT_OBIS}' requires '{CONF_NAME}' to be set")
    return value


def validate_provider_deprecation(config):
    if CONF_PROVIDER in config:
        provider = str(config[CONF_PROVIDER]).lower()
        if provider == "netznoe":
            _LOGGER.warning(
                "The 'provider: netznoe' option is deprecated and will be removed in 2026.11.0. "
                "The required custom patterns have been added automatically for this release, but you must update your configuration.\n"
                "Please remove the 'provider' key and explicitly replace it with the following:\n\n"
                "custom_patterns:\n"
                '  - pattern: "L, TSTR"\n'
                '    name: "MeterID"\n'
                '    default_obis: "0.0.96.1.0.255"\n'
                '  - pattern: "F, TDTM"\n'
                '    name: "DateTime"\n'
                '    default_obis: "0.0.1.0.0.255"\n'
            )
            patterns = config.get(CONF_CUSTOM_PATTERNS, [])

            # Ensure "L, TSTR" for MeterID is present
            if not any(p.get(CONF_PATTERN) == "L, TSTR" for p in patterns):
                patterns.append(
                    {
                        CONF_PATTERN: "L, TSTR",
                        CONF_NAME: "MeterID",
                        CONF_DEFAULT_OBIS: [0, 0, 96, 1, 0, 255],
                        CONF_PRIORITY: 0,
                    }
                )

            # Ensure "F, TDTM" for DateTime is present
            if not any(p.get(CONF_PATTERN) == "F, TDTM" for p in patterns):
                patterns.append(
                    {
                        CONF_PATTERN: "F, TDTM",
                        CONF_NAME: "DateTime",
                        CONF_DEFAULT_OBIS: [0, 0, 1, 0, 0, 255],
                        CONF_PRIORITY: 0,
                    }
                )

            config[CONF_CUSTOM_PATTERNS] = patterns
        else:
            _LOGGER.warning(
                "The 'provider' option is deprecated and will be removed in 2026.11.0. "
                "The dlms_parser library now handles quirks dynamically. "
                "Please remove this option from your configuration."
            )
    return config


CUSTOM_PATTERN_SCHEMA = cv.All(
    custom_pattern_dict,
    cv.Schema(
        {
            cv.Required(CONF_PATTERN): cv.string,
            cv.Optional(CONF_NAME): cv.string,
            cv.Optional(CONF_PRIORITY, default=0): cv.int_,
            cv.Optional(CONF_DEFAULT_OBIS): parse_obis_code_bytes,
        }
    ),
    validate_custom_pattern,
)

CONFIG_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(DlmsMeterComponent),
            cv.Optional(CONF_DECRYPTION_KEY): cv.bind_key(name="Decryption key"),
            cv.Optional(CONF_AUTH_KEY): cv.bind_key(name="Authentication key"),
            cv.Optional(CONF_CUSTOM_PATTERNS): cv.ensure_list(CUSTOM_PATTERN_SCHEMA),
            cv.Optional(CONF_SKIP_CRC, default=False): cv.boolean,
            cv.Optional(CONF_PROVIDER): cv.string,
            cv.Optional(
                CONF_RECEIVE_TIMEOUT, default="1000ms"
            ): cv.positive_time_period_milliseconds,
        }
    )
    .extend(uart.UART_DEVICE_SCHEMA)
    .extend(cv.COMPONENT_SCHEMA),
    validate_provider_deprecation,
)

FINAL_VALIDATE_SCHEMA = uart.final_validate_device_schema("dlms_meter", require_rx=True)


async def to_code(config):
    dec_key_expr = cg.RawExpression("std::nullopt")
    if dec_key := config.get(CONF_DECRYPTION_KEY):
        key_bytes = [str(int(dec_key[i : i + 2], 16)) for i in range(0, 32, 2)]
        dec_key_expr = cg.RawExpression(
            f"std::array<uint8_t, 16>{{{', '.join(key_bytes)}}}"
        )

    auth_key_expr = cg.RawExpression("std::nullopt")
    if auth_key := config.get(CONF_AUTH_KEY):
        key_bytes = [str(int(auth_key[i : i + 2], 16)) for i in range(0, 32, 2)]
        auth_key_expr = cg.RawExpression(
            f"std::array<uint8_t, 16>{{{', '.join(key_bytes)}}}"
        )

    patterns = []
    if custom_patterns := config.get(CONF_CUSTOM_PATTERNS):
        for p in custom_patterns:
            name_expr = cg.RawExpression("std::nullopt")
            if name_val := p.get(CONF_NAME):
                name_expr = name_val

            if obis_vals := p.get(CONF_DEFAULT_OBIS):
                obis_expr = cg.RawExpression(
                    f"std::array<uint8_t, 6>{{{obis_vals[0]}, {obis_vals[1]}, {obis_vals[2]}, {obis_vals[3]}, {obis_vals[4]}, {obis_vals[5]}}}"
                )
            else:
                obis_expr = cg.RawExpression("std::nullopt")

            patterns.append(
                cg.ArrayInitializer(
                    p[CONF_PATTERN],
                    name_expr,
                    p.get(CONF_PRIORITY, 0),
                    obis_expr,
                )
            )

    patterns_expr = (
        cg.ArrayInitializer(*patterns) if patterns else cg.RawExpression("{}")
    )

    var = cg.new_Pvariable(
        config[CONF_ID],
        config[CONF_RECEIVE_TIMEOUT],
        config[CONF_SKIP_CRC],
        dec_key_expr,
        auth_key_expr,
        patterns_expr,
    )

    hub_id = config[CONF_ID].id

    sensor_count = 0
    for sens_conf in CORE.config.get("sensor", []):
        if (
            sens_conf.get("platform") == "dlms_meter"
            and sens_conf.get(CONF_DLMS_METER_ID).id == hub_id
        ):
            if CONF_OBIS_CODE in sens_conf:
                sensor_count += 1
            else:
                from .sensor import NUMERIC_KEYS

                sensor_count += sum(1 for key in NUMERIC_KEYS if key in sens_conf)

    text_sensor_count = 0
    for sens_conf in CORE.config.get("text_sensor", []):
        if (
            sens_conf.get("platform") == "dlms_meter"
            and sens_conf.get(CONF_DLMS_METER_ID).id == hub_id
        ):
            if CONF_OBIS_CODE in sens_conf:
                text_sensor_count += 1
            else:
                from .text_sensor import TEXT_KEYS

                text_sensor_count += sum(1 for key in TEXT_KEYS if key in sens_conf)

    binary_sensor_count = 0
    for sens_conf in CORE.config.get("binary_sensor", []):
        if (
            sens_conf.get("platform") == "dlms_meter"
            and sens_conf.get(CONF_DLMS_METER_ID).id == hub_id
        ):
            binary_sensor_count += 1

    cg.add_define("DLMS_MAX_SENSORS", sensor_count)
    cg.add_define("DLMS_MAX_TEXT_SENSORS", text_sensor_count)
    cg.add_define("DLMS_MAX_BINARY_SENSORS", binary_sensor_count)

    await cg.register_component(var, config)
    await uart.register_uart_device(var, config)

    if CORE.is_esp32:
        esp32.add_idf_component(name="esphome/dlms_parser", ref="1.1.0")
    else:
        cg.add_library("esphome/dlms_parser", "1.1.0")



from typing import Any

import esphome.codegen as cg
import esphome.config_validation as cv
from esphome.const import CONF_MAX_VALUE, CONF_MIN_VALUE

from . import generate, schema

CONF_AUTO_MIN_VALUE = "auto_min_value"
CONF_AUTO_MAX_VALUE = "auto_max_value"

OpenthermInput = generate.opentherm_ns.class_("OpenthermInput")


def validate_min_value_less_than_max_value(conf):
    if (
        CONF_MIN_VALUE in conf
        and CONF_MAX_VALUE in conf
        and conf[CONF_MIN_VALUE] > conf[CONF_MAX_VALUE]
    ):
        raise cv.Invalid(f"{CONF_MIN_VALUE} must be less than {CONF_MAX_VALUE}")
    return conf


def input_schema(entity: schema.InputSchema) -> cv.Schema:
    result = cv.Schema(
        {
            cv.Optional(CONF_MIN_VALUE, entity.range[0]): cv.float_range(
                entity.range[0], entity.range[1]
            ),
            cv.Optional(CONF_MAX_VALUE, entity.range[1]): cv.float_range(
                entity.range[0], entity.range[1]
            ),
        }
    )
    result = result.add_extra(validate_min_value_less_than_max_value)
    if entity.auto_min_value is not None:
        result = result.extend({cv.Optional(CONF_AUTO_MIN_VALUE, False): cv.boolean})
    if entity.auto_max_value is not None:
        result = result.extend({cv.Optional(CONF_AUTO_MAX_VALUE, False): cv.boolean})

    return result


def generate_setters(entity: cg.MockObj, conf: dict[str, Any]) -> None:
    generate.add_property_set(entity, CONF_MIN_VALUE, conf)
    generate.add_property_set(entity, CONF_MAX_VALUE, conf)
    generate.add_property_set(entity, CONF_AUTO_MIN_VALUE, conf)
    generate.add_property_set(entity, CONF_AUTO_MAX_VALUE, conf)



import esphome.codegen as cg
from esphome.components.ota import BASE_OTA_SCHEMA, OTAComponent, ota_to_code
from esphome.components.zephyr import (
    zephyr_add_cdc_acm,
    zephyr_add_overlay,
    zephyr_add_prj_conf,
    zephyr_data,
)
from esphome.components.zephyr.const import (
    BOOTLOADER_MCUBOOT,
    KEY_BOOTLOADER,
    KEY_SYSBUILD,
)
import esphome.config_validation as cv
from esphome.const import (
    CONF_HARDWARE_UART,
    CONF_ID,
    KEY_CORE,
    KEY_FRAMEWORK_VERSION,
    Framework,
)
from esphome.core import CORE, coroutine_with_priority
from esphome.coroutine import CoroPriority
from esphome.types import ConfigType

CODEOWNERS = ["@tomaszduda23"]
DEPENDENCIES = ["zephyr"]

ZephyrMcumgrOTAComponent = cg.esphome_ns.namespace("zephyr_mcumgr").class_(
    "OTAComponent", OTAComponent
)

CONF_BLE = "ble"
CONF_TRANSPORT = "transport"


def _validate_transport(conf: ConfigType) -> ConfigType:
    transport = conf[CONF_TRANSPORT]
    if transport[CONF_BLE] or CONF_HARDWARE_UART in transport:
        return conf
    raise cv.Invalid(
        f"At least one transport protocol has to be enabled. Set '{CONF_BLE}: true' or '{CONF_HARDWARE_UART}'"
    )


UARTS = {
    "CDC": ("cdc_acm_uart0", 0),
    "CDC1": ("cdc_acm_uart1", 1),
    "UART0": ("uart0", -1),
    "UART1": ("uart1", -1),
}


CONFIG_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(ZephyrMcumgrOTAComponent),
            cv.Optional(CONF_TRANSPORT, default={CONF_BLE: True}): cv.Schema(
                {
                    cv.Optional(CONF_BLE, default=False): cv.boolean,
                    cv.Optional(
                        CONF_HARDWARE_UART,
                    ): cv.one_of(*UARTS, upper=True),
                }
            ),
        }
    )
    .extend(BASE_OTA_SCHEMA)
    .extend(cv.COMPONENT_SCHEMA),
    _validate_transport,
    cv.only_with_framework(Framework.ZEPHYR),
)


def _validate_mcumgr_bootloader(config: ConfigType) -> None:
    bootloader = zephyr_data()[KEY_BOOTLOADER]
    if bootloader != BOOTLOADER_MCUBOOT:
        raise cv.Invalid(f"'{bootloader}' bootloader does not support OTA")


KEY_ZEPHYR_BLE_SERVER = "zephyr_ble_server"


def _validate_ble_server(config: ConfigType) -> None:
    if (
        config[CONF_TRANSPORT][CONF_BLE]
        and KEY_ZEPHYR_BLE_SERVER not in CORE.loaded_integrations
    ):
        raise cv.Invalid(f"'{KEY_ZEPHYR_BLE_SERVER}' component is required for BLE OTA")


def _final_validate(config: ConfigType) -> None:
    _validate_mcumgr_bootloader(config)
    _validate_ble_server(config)


FINAL_VALIDATE_SCHEMA = _final_validate


@coroutine_with_priority(CoroPriority.OTA_UPDATES)
async def to_code(config: ConfigType) -> None:
    var = cg.new_Pvariable(config[CONF_ID])
    await ota_to_code(var, config)

    await cg.register_component(var, config)

    zephyr_add_prj_conf("NET_BUF", True)
    zephyr_add_prj_conf("ZCBOR", True)
    zephyr_add_prj_conf("MCUMGR", True)

    zephyr_add_prj_conf("MCUMGR_GRP_IMG", True)

    zephyr_add_prj_conf("IMG_MANAGER", True)
    zephyr_add_prj_conf("STREAM_FLASH", True)
    zephyr_add_prj_conf("FLASH_MAP", True)
    zephyr_add_prj_conf("FLASH", True)

    zephyr_add_prj_conf("IMG_ERASE_PROGRESSIVELY", True)

    zephyr_add_prj_conf("BOOTLOADER_MCUBOOT", True)

    zephyr_add_prj_conf("MCUMGR_MGMT_NOTIFICATION_HOOKS", True)
    zephyr_add_prj_conf("MCUMGR_GRP_IMG_STATUS_HOOKS", True)
    zephyr_add_prj_conf("MCUMGR_GRP_IMG_UPLOAD_CHECK_HOOK", True)
    transport = config[CONF_TRANSPORT]
    if transport[CONF_BLE]:
        zephyr_add_prj_conf("MCUMGR_TRANSPORT_BT", True)
        zephyr_add_prj_conf("MCUMGR_TRANSPORT_BT_REASSEMBLY", True)

        zephyr_add_prj_conf("MCUMGR_GRP_OS", True)
        zephyr_add_prj_conf("MCUMGR_GRP_OS_MCUMGR_PARAMS", True)

        zephyr_add_prj_conf("NCS_SAMPLE_MCUMGR_BT_OTA_DFU_SPEEDUP", True)
    if CONF_HARDWARE_UART in transport:
        uart = UARTS[transport[CONF_HARDWARE_UART]]
        uart_name = uart[0]
        cdc_id = uart[1]
        if cdc_id >= 0:
            zephyr_add_cdc_acm(config, cdc_id)
        zephyr_add_prj_conf("MCUMGR_TRANSPORT_UART", True)
        zephyr_add_prj_conf("BASE64", True)
        zephyr_add_prj_conf("CONSOLE", True)
        zephyr_add_overlay(
            f"""
                / {{
                    chosen {{
                        zephyr,uart-mcumgr = &{uart_name};
                    }};
                }};
                """
        )
    framework_ver = CORE.data[KEY_CORE][KEY_FRAMEWORK_VERSION]
    if framework_ver >= cv.Version(2, 9, 2):
        zephyr_data()[KEY_SYSBUILD] = True



from collections import ChainMap
import logging
from typing import Any

import esphome
from esphome import core
from esphome.config_helpers import Extend, Remove, merge_config, merge_dicts_ordered
import esphome.config_validation as cv
from esphome.const import CONF_SUBSTITUTIONS, VALID_SUBSTITUTIONS_CHARACTERS
from esphome.types import ConfigType
from esphome.util import OrderedDict
from esphome.yaml_util import (
    ConfigContext,
    DocumentPath,
    ESPHomeDataBase,
    ESPLiteralValue,
    IncludeFile,
    format_path,
    make_data_base,
)

from .jinja import Jinja, JinjaError, Missing, Resolver, UndefinedError, has_jinja

CODEOWNERS = ["@esphome/core"]
_LOGGER = logging.getLogger(__name__)

ContextVars = ChainMap[str, Any]
ErrList = list[tuple[UndefinedError, DocumentPath, Any]]

# Module-level instance is safe: context_vars is passed per-call, and context_trace
# is stack-saved/restored within expand(). Not thread-safe — only use from one thread.
jinja = Jinja()


def raise_first_undefined(
    errors: ErrList,
    context_label: str,
) -> None:
    """If *errors* is non-empty, raise ``cv.Invalid`` for the first undefined variable.

    The raised error names the missing variable and its location in the include
    stack. Only the first error is surfaced; the user will re-run after fixing it
    and any remaining undefined variables will be reported then.

    ``context_label`` is the noun describing where the undefined variable
    appeared (e.g. ``"package definition"``).
    """
    if not errors:
        return
    err, err_path, err_value = errors[0]
    if len(errors) > 1:
        # Log any further undefined variables so debug-level output covers
        # the full set, even though only the first is surfaced to the user.
        extras = ", ".join(
            f"{e.message} at '{'->'.join(str(p) for p in p_path)}'"
            for e, p_path, _ in errors[1:]
        )
        _LOGGER.debug("Additional undefined variables in %s: %s", context_label, extras)
    raise cv.Invalid(
        f"Undefined variable in {context_label}: {err.message}\n{format_path(err_path, err_value)}"
    )


def validate_substitution_key(value: Any) -> str:
    """Validate and normalize a substitution key, stripping a leading ``$`` if present."""
    value = cv.string(value)
    if not value:
        raise cv.Invalid("Substitution key must not be empty")
    if value[0] == "$":
        value = value[1:]
    if not value:
        raise cv.Invalid("Substitution key must not be empty")
    if value[0].isdigit():
        raise cv.Invalid("First character in substitutions cannot be a digit.")
    for char in value:
        if char not in VALID_SUBSTITUTIONS_CHARACTERS:
            raise cv.Invalid(
                f"Substitution must only consist of upper/lowercase characters,"
                f" the underscore and numbers."
                f" The character '{char}' cannot be used"
            )
    return value


CONFIG_SCHEMA = cv.Schema(
    {
        validate_substitution_key: object,
    }
)


async def to_code(config: ConfigType) -> None:
    """No runtime code generation needed — substitutions are resolved at config time."""


def _restore_data_base(value: Any, orig_value: ESPHomeDataBase) -> ESPHomeDataBase:
    """This function restores ESPHomeDataBase metadata held by the original string.
    This is needed because during jinja evaluation, strings can be replaced by other types,
    but we want to keep the original metadata for error reporting and source mapping.
    For example, if a substitution replaces a string with a dictionary, we want that items
    in the dictionary to still point to the original document location
    """
    if isinstance(value, ESPHomeDataBase):
        return value
    if isinstance(value, dict):
        return {
            _restore_data_base(k, orig_value): _restore_data_base(v, orig_value)
            for k, v in value.items()
        }
    if isinstance(value, list):
        return [_restore_data_base(v, orig_value) for v in value]
    if isinstance(value, str):
        return make_data_base(value, orig_value)
    return value


def _resolve_var(name: str, context_vars: ContextVars) -> Any:
    """Look up a substitution variable, falling back to the resolver callback."""
    sub = context_vars.get(name, Missing)
    if sub is Missing:
        resolver = context_vars.get(Resolver)
        if resolver:
            sub = resolver(name)
    return sub


def _handle_undefined(
    err: UndefinedError,
    path: DocumentPath,
    value: Any,
    strict_undefined: bool,
    errors: ErrList | None,
) -> None:
    """Handle an undefined variable.

    In strict mode, raises immediately. Otherwise, appends to the errors
    list for deferred warning at the end of the substitution pass.
    """
    if strict_undefined:
        raise err
    if errors is not None:
        errors.append((err, path, value))


def _expand_substitutions(
    value: str,
    path: DocumentPath,
    context_vars: ContextVars,
    strict_undefined: bool,
    errors: ErrList | None,
) -> Any:
    """Expand ``$var``, ``${var}``, and Jinja expressions in a string.

    Works in two phases:

    1. **Simple substitution** — scan for ``$name`` / ``${name}`` tokens
       and replace them with the value from *context_vars*.  If the token
       spans the entire string, return the raw value (preserving type).
    2. **Jinja evaluation** — if the result still contains Jinja syntax
       (e.g. ``${a * b}``), render it through the Jinja engine with the
       full *context_vars* as template variables.

    Returns the expanded value (may be a non-string type) or the
    original *value* unchanged if there is nothing to substitute.
    """
    if "$" not in value:
        return value

    orig_value = value

    # Phase 1: Replace $var and ${var} references
    search_pos = 0
    while (m := cv.VARIABLE_PROG.search(value, search_pos)) is not None:
        match_start, match_end = m.span(0)
        name: str = m.group(1)
        if name.startswith("{") and name.endswith("}"):
            name = name[1:-1]
        sub = _resolve_var(name, context_vars)
        if sub is Missing:
            _handle_undefined(
                err=UndefinedError(f"'{name}' is undefined"),
                path=path,
                value=value,
                strict_undefined=strict_undefined,
                errors=errors,
            )
            search_pos = match_end
            continue

        if match_start == 0 and match_end == len(value):
            # The variable spans the whole expression, e.g., "${varName}".
            # Return its resolved value directly to conserve its type.
            value = sub
            break

        tail = value[match_end:]
        value = value[:match_start] + str(sub)
        search_pos = len(value)
        value += tail

    # Phase 2: Evaluate any remaining jinja expressions (e.g., "${a * b}")
    if isinstance(value, str) and has_jinja(value):
        try:
            value = jinja.expand(value, context_vars)
        except UndefinedError as err:
            _handle_undefined(
                err=err,
                path=path,
                value=value,
                strict_undefined=strict_undefined,
                errors=errors,
            )
        except JinjaError as err:
            raise cv.Invalid(
                f"{err.error_name()} Error evaluating jinja expression"
                f" '{value}': {str(err.parent())}."
                f"\nEvaluation stack: (most recent evaluation last)"
                f"\n{err.stack_trace_str()}"
                f"\nRelevant context:\n{err.context_trace_str()}"
                f"\n{format_path(path, orig_value)}",
                path,
            ) from err
        else:
            if isinstance(orig_value, ESPHomeDataBase):
                value = _restore_data_base(value, orig_value)

    # orig_value can also already be a lambda with esp_range info, and only
    # a plain string is sent in orig_value
    if isinstance(orig_value, ESPHomeDataBase):
        # even though string can get larger or smaller, the range should point
        # to original document marks
        value = make_data_base(value, orig_value)

    return value


def _push_context(
    local_vars: dict[str, Any],
    parent_context: ContextVars,
    errors: ErrList | None = None,
) -> tuple[ContextVars, dict[str, Any]]:
    """Resolve local_vars and layer them on top of parent_context.

    Returns ``(child_context, resolved_vars)`` where *child_context* is a
    new :class:`ChainMap` whose front map is *resolved_vars* (an
    :class:`OrderedDict` of successfully-resolved variables).

    Variables may reference each other (e.g. ``b: ${a + 1}``).
    Dependencies are resolved recursively via a *resolver* callback
    that Jinja invokes on cache-miss.  If vars are already in
    dependency order, the loop iterates exactly once per variable.

    The ChainMap stack used during resolution is::

        resolver_context  →  resolved_vars  →  parent maps …
              ↑                    ↑
        holds Resolver       filled as vars
        callback               are resolved
    """
    # Vars still waiting to be resolved — popped one-by-one by resolve().
    unresolved_vars = local_vars.copy()
    # Accumulates resolved values in dependency order; becomes the front
    # map of the returned child context so later lookups find them first.
    resolved_vars = OrderedDict()
    # The context callees will search: resolved_vars (initially empty)
    # shadowing whatever the parent already provides.
    context_vars = parent_context.new_child(resolved_vars)

    # Vars that failed resolution (missing or circular references).
    # Maps name → (original_value, cause_error) for deferred warnings.
    unresolvables: dict[str, tuple[Any, UndefinedError]] = {}

    # One extra child layer so the Resolver callback lives in its own
    # map and doesn't pollute resolved_vars.
    resolver_context = context_vars.new_child()

    def resolve(key: str) -> Any:
        """Resolve a variable, recursively resolving any dependencies it references."""
        value = unresolved_vars.pop(key, Missing)
        if value is Missing:
            # Either already resolved (in resolved_vars) or currently being
            # resolved (self-reference from inside a dict-valued substitution).
            # Returning what we have lets sibling references inside a dict
            # value, e.g. ``${device.manufacturer}`` inside ``device.name``,
            # see literal sibling values during their own resolution.
            return resolved_vars.get(key, Missing)
        if isinstance(value, dict):
            # Dict-valued substitutions form a namespace; eagerly publish the
            # original mapping so its members can reference each other while
            # the dict's own substitution pass is still running. The entry is
            # replaced with the fully-substituted dict once recursion returns.
            resolved_vars[key] = value
        try:
            value = substitute(value, [], resolver_context, True)
        except UndefinedError as err:
            unresolvables[key] = (value, err)
            return Missing
        resolved_vars[key] = value
        return value

    # Set up the resolver for use during substitution
    resolver_context[Resolver] = resolve

    # Resolve all variables, recursively resolving dependencies as needed.
    # Each call to resolve() resolves that variable and any variables it depends on.
    while unresolved_vars:
        resolve(next(iter(unresolved_vars)))

    for name, (value, cause) in unresolvables.items():
        resolved_vars[name] = value
        if errors is not None:
            _handle_undefined(
                err=UndefinedError(
                    f"Could not resolve substitution variable '{name}': {cause}"
                ),
                path=["substitutions", name],
                value=value,
                strict_undefined=False,
                errors=errors,
            )

    return context_vars, resolved_vars


def push_context(
    config_node: Any,
    parent_context: ContextVars,
    errors: ErrList | None = None,
) -> ContextVars:
    """Returns the context vars this config node must be evaluated with."""
    if isinstance(config_node, ConfigContext):
        return _push_context(config_node.vars, parent_context, errors)[0]

    # This node does not define any vars itself, so just return parent context
    return parent_context


def resolve_include(
    include: IncludeFile,
    path: DocumentPath,
    context_vars: ContextVars,
    strict_undefined: bool = True,
    errors: ErrList | None = None,
) -> Any:
    """Resolve an include, substituting the filename if needed.

    Note: no path-traversal validation is performed on the resolved filename.
    A substitution that resolves to an absolute path will bypass the parent
    directory (Path.__truediv__ ignores the left operand for absolute paths).
    ESPHome's trust model assumes the config author controls all substitution
    values (including command-line substitutions), so path restrictions are
    an explicit non-goal here.
    """
    original = include.file
    original_str = str(original)
    filename = str(
        _expand_substitutions(
            original_str, path + ["file"], context_vars, strict_undefined, errors
        )
    )
    substituted = filename != original_str
    if substituted:
        include = IncludeFile(
            include.parent_file, filename, include.vars, include.yaml_loader
        )
    try:
        return include.load()
    except esphome.core.EsphomeError as err:
        resolved = f" (expanded from '{original}')" if substituted else ""
        raise cv.Invalid(
            f"Error including file '{filename}'{resolved}: {err}"
            f"\n{format_path(path, original)}",
            path + [f"<{filename}>"],
        ) from err


def _substitute_include(
    include: IncludeFile,
    path: DocumentPath,
    context_vars: ContextVars,
    strict_undefined: bool,
    errors: ErrList | None,
) -> Any:
    """Resolve an include and substitute its content."""
    content = resolve_include(include, path, context_vars, strict_undefined, errors)
    return substitute(content, path, context_vars, strict_undefined, errors)


def substitute(
    item: Any,
    path: DocumentPath,
    parent_context: ContextVars,
    strict_undefined: bool,
    errors: ErrList | None = None,
) -> Any:
    """Returns a recursively substituted version of `item`."""

    if isinstance(item, ESPLiteralValue):
        return item  # do not substitute inside literal blocks

    # Push the current item's context onto the context stack
    context_vars = push_context(item, parent_context, errors)

    result = item

    if isinstance(item, list):
        result = [
            substitute(it, path + [i], context_vars, strict_undefined, errors)
            for i, it in enumerate(item)
        ]

    elif isinstance(item, dict):
        result = OrderedDict()
        for k, v in item.items():
            v = substitute(v, path + [k], context_vars, strict_undefined, errors)
            k = substitute(k, path + [k], context_vars, strict_undefined, errors)
            result[k] = merge_config(result.get(k), v)

    elif isinstance(item, str):
        result = _expand_substitutions(
            item, path, context_vars, strict_undefined, errors
        )

    elif isinstance(item, (core.Lambda, Extend, Remove)) and item.value:
        value = _expand_substitutions(
            item.value, path, context_vars, strict_undefined, errors
        )
        if item.value != value:
            result = type(item)(value)

    elif isinstance(item, IncludeFile):
        result = _substitute_include(item, path, context_vars, strict_undefined, errors)

    if isinstance(item, ESPHomeDataBase):
        result = make_data_base(result, item)
    return result


def _warn_unresolved_variables(errors: ErrList) -> None:
    """Log warnings for unresolved substitution variables, skipping password fields."""
    for err, path, expression in errors:
        if "password" in path:
            continue
        _LOGGER.warning(
            "The string '%s' looks like an expression,"
            " but could not resolve all the variables: %s\n%s",
            expression,
            err.message,
            format_path(path, expression),
        )


def resolve_substitutions_block(
    substitutions: Any,
    command_line_substitutions: dict[str, Any] | None,
) -> dict[str, Any]:
    """Resolve a deferred ``substitutions: !include file.yaml`` and validate the shape.

    The caller is responsible for wrapping the call in
    ``cv.prepend_path(CONF_SUBSTITUTIONS)`` for error reporting.
    ``command_line_substitutions`` seeds the filename context so
    ``substitutions: !include ${var}.yaml`` can reference CLI-provided vars.
    """
    if isinstance(substitutions, IncludeFile):
        # Single-shot resolution — matches ``_walk_packages`` for the
        # ``packages: !include`` entry point.  Chained includes (an include that
        # itself loads another ``!include`` at the top level) are not supported.
        substitutions = resolve_include(
            substitutions,
            [],
            ContextVars(command_line_substitutions or {}),
            strict_undefined=False,
        )
    if not isinstance(substitutions, dict):
        raise cv.Invalid(
            f"Substitutions must be a key to value mapping, got {type(substitutions)}"
        )
    return substitutions


def do_substitution_pass(
    config: OrderedDict, command_line_substitutions: dict[str, Any] | None = None
) -> OrderedDict:
    """Run the substitution pass over the entire config.

    Extracts the ``substitutions:`` block, merges in any command-line
    overrides, resolves inter-variable dependencies, then walks the
    config tree replacing all ``$var`` / ``${expr}`` references.
    Returns a new config dict with resolved substitutions
    restored at the front.
    """
    # Extract substitutions from config, overriding with substitutions coming from command line:
    # Use merge_dicts_ordered to preserve OrderedDict type for move_to_end()
    substitutions = config.pop(CONF_SUBSTITUTIONS, {})
    with cv.prepend_path(CONF_SUBSTITUTIONS):
        substitutions = resolve_substitutions_block(
            substitutions, command_line_substitutions
        )
        substitutions = merge_dicts_ordered(
            substitutions, command_line_substitutions or {}
        )

        replace_keys: list[tuple[str, str]] = []
        for key in substitutions:
            with cv.prepend_path(key):
                sub = validate_substitution_key(key)
                if sub != key:
                    replace_keys.append((key, sub))
        for old, new in replace_keys:
            substitutions[new] = substitutions[old]
            del substitutions[old]

    errors: ErrList = []  # Collect undefined errors during substitution
    parent_context, substitutions = _push_context(substitutions, ContextVars(), errors)

    config = substitute(config, [], parent_context, False, errors)

    if errors:
        _warn_unresolved_variables(errors)

    # Restore substitutions to front of dict for readability
    if substitutions:
        config[CONF_SUBSTITUTIONS] = substitutions
        config.move_to_end(CONF_SUBSTITUTIONS, last=False)
    return config



import esphome.codegen as cg
import esphome.config_validation as cv
from esphome.const import CONF_ENABLE_ON_BOOT, CONF_ID
from esphome.types import ConfigType

DEPENDENCIES = ["rp2"]
CODEOWNERS = ["@bdraco"]

rp2040_ble_ns = cg.esphome_ns.namespace("rp2040_ble")
RP2040BLE = rp2040_ble_ns.class_("RP2040BLE", cg.Component)

CONFIG_SCHEMA = cv.Schema(
    {
        cv.GenerateID(): cv.declare_id(RP2040BLE),
        cv.Optional(CONF_ENABLE_ON_BOOT, default=True): cv.boolean,
    }
).extend(cv.COMPONENT_SCHEMA)


async def to_code(config: ConfigType) -> None:
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)

    cg.add(var.set_enable_on_boot(config[CONF_ENABLE_ON_BOOT]))

    # Enable Bluetooth in the arduino-pico build
    # This switches linking from liblwip.a to liblwip-bt.a and defines
    # ENABLE_CLASSIC, ENABLE_BLE, CYW43_ENABLE_BLUETOOTH
    cg.add_build_flag("-DPIO_FRAMEWORK_ARDUINO_ENABLE_BLUETOOTH")

    cg.add_define("USE_RP2040_BLE")



from typing import Any

import esphome.config_validation as cv
from esphome.const import CONF_DEVICE, CONF_ID, CONF_TYPE
from esphome.core import CORE

from .const import (
    CONF_MAX_EP_NUMBER,
    CONF_REPORT,
    CONF_USE_DEVICE_TYPE,
    KEY_ZIGBEE,
    REPORT,
)
from .const_esp32 import (
    CONF_ATTRIBUTE_ID,
    CONF_ATTRIBUTES,
    CONF_CLUSTERS,
    DEVICE_TYPE,
    KEY_ZIGBEE_EP,
    KEY_ZIGBEE_EP_NO_NUM,
    ROLE,
)

# endpoint configs:
ep_configs: dict[str, dict[str, Any]] = {
    "binary_input": {
        DEVICE_TYPE: "SIMPLE_SENSOR",
        CONF_CLUSTERS: [
            {
                CONF_ID: "BINARY_INPUT",
                ROLE: "SERVER",
                CONF_ATTRIBUTES: [
                    {
                        CONF_ATTRIBUTE_ID: 0x55,
                        CONF_TYPE: "BOOL",
                        CONF_REPORT: cv.enum(REPORT, lower=True)("default"),
                        CONF_DEVICE: None,
                    },
                    {
                        CONF_ATTRIBUTE_ID: 0x51,
                        CONF_TYPE: "BOOL",
                    },
                    {
                        CONF_ATTRIBUTE_ID: 0x6F,
                        CONF_TYPE: "MAP8",
                    },
                    {
                        CONF_ATTRIBUTE_ID: 0x1C,
                        CONF_TYPE: "STRING",
                    },
                ],
            },
        ],
    },
    "analog_input": {
        CONF_CLUSTERS: [
            {
                CONF_ID: "ANALOG_INPUT",
                ROLE: "SERVER",
                CONF_ATTRIBUTES: [
                    {
                        CONF_ATTRIBUTE_ID: 0x55,
                        CONF_TYPE: "SINGLE",
                        CONF_REPORT: cv.enum(REPORT, lower=True)("default"),
                        CONF_DEVICE: None,
                    },
                    {
                        CONF_ATTRIBUTE_ID: 0x51,
                        CONF_TYPE: "BOOL",
                    },
                    {
                        CONF_ATTRIBUTE_ID: 0x6F,
                        CONF_TYPE: "MAP8",
                    },
                    {
                        CONF_ATTRIBUTE_ID: 0x1C,
                        CONF_TYPE: "STRING",
                    },
                ],
            },
        ],
    },
}


def get_next_ep_num(eps: list[int]) -> int:
    try:
        ep_num = [i for i in range(1, CONF_MAX_EP_NUMBER + 1) if i not in eps][0]
        eps.append(ep_num)
    except IndexError as e:
        raise cv.Invalid(
            f"Too many devices. Zigbee can define only {CONF_MAX_EP_NUMBER} endpoints."
        ) from e
    return ep_num


def compare_clusters(
    existing_ep: dict[str, Any],
    ep: dict[str, Any],
) -> tuple[str | int, str] | None:
    existing_clusters = [(cl[CONF_ID], cl[ROLE]) for cl in existing_ep[CONF_CLUSTERS]]
    for cl in [(cl[CONF_ID], cl[ROLE]) for cl in ep[CONF_CLUSTERS]]:
        if cl in existing_clusters:
            return cl
    return None


def merge_endpoints(
    existing_ep: dict[str, Any],
    ep: dict[str, Any],
    use_type: bool | None,
) -> bool:
    if compare_clusters(existing_ep, ep):
        return False
    if (
        ep.get(DEVICE_TYPE)
        and existing_ep.get(DEVICE_TYPE)
        and ep.get(DEVICE_TYPE) != existing_ep.get(DEVICE_TYPE)
    ):
        return False
    if (
        ep.get(DEVICE_TYPE)
        and not existing_ep.get(DEVICE_TYPE)
        and existing_ep.get(CONF_USE_DEVICE_TYPE)
    ):
        return False
    if existing_ep.get(DEVICE_TYPE) and not ep.get(DEVICE_TYPE) and use_type:
        return False
    if use_type:
        existing_ep[CONF_USE_DEVICE_TYPE] = use_type
    if ep.get(DEVICE_TYPE):
        existing_ep[DEVICE_TYPE] = ep[DEVICE_TYPE]
    existing_ep[CONF_CLUSTERS].extend(ep[CONF_CLUSTERS])
    return True


def validate_endpoints(ep_dict: dict[int, dict]) -> None:
    for num, ep in ep_dict.items():
        types_dict = ep.get(CONF_USE_DEVICE_TYPE)
        if not types_dict:
            continue
        if len(types_dict) == 1:
            ep[DEVICE_TYPE] = list(types_dict.keys())[0]
            del ep[CONF_USE_DEVICE_TYPE]
            continue
        types_list = [t[0] for t in types_dict.items() if t[1]]
        if len(types_list) > 1:
            raise cv.Invalid(
                f"There is more than one component with endpoint: {num} and {CONF_USE_DEVICE_TYPE}: True"
            )
        if not types_list:
            raise cv.Invalid(
                f"Multiple device types on endpoint: {num}. Set {CONF_USE_DEVICE_TYPE}: True on one component."
            )
        ep[DEVICE_TYPE] = types_list[0]
        del ep[CONF_USE_DEVICE_TYPE]


def create_ep(router: bool) -> None:
    zb_data = CORE.data.setdefault(KEY_ZIGBEE, {})
    ep_dict: dict[int, dict] = zb_data.setdefault(KEY_ZIGBEE_EP, {})
    ep_list: list[dict] = zb_data.setdefault(KEY_ZIGBEE_EP_NO_NUM, [])
    validate_endpoints(ep_dict)
    # create dummy endpoint if list is empty
    if not ep_dict and not ep_list:
        ep_type = "CUSTOM_ATTR"
        if router:
            ep_type = "RANGE_EXTENDER"
        ep_dict[1] = {DEVICE_TYPE: ep_type}
    if ep_list:
        # merge endpoint with different clusters
        ep_list_new: list[dict] = []
        for ep in ep_list:
            added = False
            for existing_ep in ep_list_new:
                if merge_endpoints(existing_ep, ep, ep.get(CONF_USE_DEVICE_TYPE)):
                    added = True
                    break
            if not added:
                ep_list_new.append(ep)

        # Add endpoints with no number to the endpoint dict with a new number
        eps = list(ep_dict.keys())
        for ep in ep_list_new:
            ep_num = get_next_ep_num(eps)
            ep_dict[ep_num] = ep

        # clear list so that it is not processed again
        del zb_data[KEY_ZIGBEE_EP_NO_NUM]

    # Add default device type to endpoints that have none
    for ep in ep_dict.values():
        if not ep.get(DEVICE_TYPE):
            ep[DEVICE_TYPE] = "CUSTOM_ATTR"


def add_ep(ep: dict[str, Any], ep_num: int | None, use_type: bool | None) -> None:
    zb_data = CORE.data.setdefault(KEY_ZIGBEE, {})
    if use_type is False:
        ep.pop(DEVICE_TYPE, None)
    if ep_num is None:
        if use_type:
            ep[CONF_USE_DEVICE_TYPE] = use_type
        ep_list: list[dict] = zb_data.setdefault(KEY_ZIGBEE_EP_NO_NUM, [])
        ep_list.append(ep)
    else:
        ep_dict: dict[int, dict] = zb_data.setdefault(KEY_ZIGBEE_EP, {})
        if ep_num in ep_dict:
            # check if the existing endpoint has same clusters
            existing_ep = ep_dict[ep_num]
            if cl := compare_clusters(
                existing_ep,
                ep,
            ):
                raise cv.Invalid(
                    f"Endpoint {ep_num} has more than one cluster with cluster id {cl[0]} and role {cl[1]}."
                )
            if ep.get(DEVICE_TYPE) or use_type:
                types_dict = existing_ep.setdefault(CONF_USE_DEVICE_TYPE, {})
                if not types_dict.get(ep.get(DEVICE_TYPE)) or use_type:
                    types_dict[ep.get(DEVICE_TYPE)] = use_type
            existing_ep[CONF_CLUSTERS].extend(ep[CONF_CLUSTERS])
        else:
            if use_type or ep.get(DEVICE_TYPE):
                ep[CONF_USE_DEVICE_TYPE] = {ep.get(DEVICE_TYPE): use_type}
            ep_dict[ep_num] = ep



from collections.abc import Callable
import difflib

import esphome.codegen as cg
from esphome.components.const import KEY_METADATA
import esphome.config_validation as cv
from esphome.const import CONF_FROM, CONF_ID, CONF_TO
from esphome.core import CORE, ID
from esphome.cpp_generator import (
    MockObj,
    MockObjClass,
    VariableDeclarationExpression,
    add_global,
)
from esphome.loader import get_component

CODEOWNERS = ["@clydebarrow"]
MULTI_CONF = True
DOMAIN = "mapping"

mapping_ns = cg.esphome_ns.namespace("mapping")
mapping_class = mapping_ns.class_("Mapping")

CONF_DEFAULT_VALUE = "default_value"
CONF_ENTRIES = "entries"
CONF_CLASS = "class"


class IndexType:
    """
    Represents a type of index in a map.
    """

    def __init__(
        self, validator: Callable, data_type: MockObj, conversion: Callable = None
    ) -> None:
        self.validator = validator
        self.data_type = data_type
        self.conversion = conversion

    async def convert_value(self, value):
        if self.conversion:
            return self.conversion(value)
        return await cg.get_variable(value)


INDEX_TYPES = {
    "int": IndexType(cv.int_, cg.int_, int),
    "string": IndexType(
        cv.string,
        cg.std_string,
        str,
    ),
}


class MappingMetaData:
    def __init__(self, from_: IndexType, to_: IndexType) -> None:
        self.from_ = from_
        self.to_ = to_


def to_schema(value):
    """
    Generate a schema for the 'to' field of a map. This can be either one of the index types or a class name.
    :param value:
    :return:
    """
    return cv.Any(
        cv.one_of(*INDEX_TYPES, lower=True),
        cv.one_of(*CORE.id_classes.keys()),
    )(value)


BASE_SCHEMA = cv.Schema(
    {
        cv.Required(CONF_ID): cv.declare_id(mapping_class),
        cv.Required(CONF_FROM): cv.one_of(*INDEX_TYPES, lower=True),
        cv.Required(CONF_TO): cv.string,
    },
    extra=cv.ALLOW_EXTRA,
)


def get_object_type(to_) -> MockObjClass | None:
    """
    Get the object type from a string. Possible formats:
       xxx The name of a component which defines INSTANCE_TYPE
       esphome::xxx::yyy A C++ class name defined in a component
       xxx::yyy A C++ class name defined in a component
       yyy A C++ class name defined in the core
    """

    if cls := CORE.id_classes.get(to_):
        return cls
    if cls := CORE.id_classes.get(to_.removeprefix("esphome::")):
        return cls
    # get_component will throw a wobbly if we don't check this first.
    if "." in to_:
        return None
    if component := get_component(to_):
        return component.instance_type
    return None


def get_all_mapping_metadata() -> dict[str, MappingMetaData]:
    """Get all mapping metadata."""
    return CORE.data.setdefault(DOMAIN, {}).setdefault(KEY_METADATA, {})


def get_mapping_metadata(mapping_id: str) -> MappingMetaData:
    """Get mapping metadata by ID for use by other components."""
    return get_all_mapping_metadata()[mapping_id]


def add_metadata(
    mapping_id: ID,
    from_: IndexType,
    to_: IndexType,
) -> None:
    get_all_mapping_metadata()[mapping_id.id] = MappingMetaData(from_, to_)


def map_schema(config):
    config = BASE_SCHEMA(config)
    if CONF_ENTRIES not in config or not isinstance(config[CONF_ENTRIES], dict):
        raise cv.Invalid("an entries dictionary is required for a mapping")
    entries = config[CONF_ENTRIES]
    if len(entries) == 0:
        raise cv.Invalid("A mapping must have at least one entry")
    to_ = config[CONF_TO]
    if to_ in INDEX_TYPES:
        value_type = INDEX_TYPES[to_]
    else:
        object_type = get_object_type(to_)
        if object_type is None:
            matches = difflib.get_close_matches(to_, CORE.id_classes)
            raise cv.Invalid(
                f"No known mappable class name matches '{to_}'; did you mean one of {', '.join(matches)}?"
            )
        validator = cv.use_id(object_type)
        value_type = IndexType(validator, object_type)
    config[CONF_ENTRIES] = {k: value_type.validator(v) for k, v in entries.items()}
    if (default_value := config.get(CONF_DEFAULT_VALUE)) is not None:
        config[CONF_DEFAULT_VALUE] = value_type.validator(default_value)
    unexpected_keys = config.keys() - {
        CONF_ENTRIES,
        CONF_TO,
        CONF_FROM,
        CONF_ID,
        CONF_DEFAULT_VALUE,
    }
    if unexpected_keys:
        errors = [
            cv.Invalid(f"Unexpected key '{k}'", path=[k]) for k in unexpected_keys
        ]
        raise cv.MultipleInvalid(errors)

    add_metadata(config[CONF_ID], INDEX_TYPES[config[CONF_FROM]], value_type)
    return config


CONFIG_SCHEMA = map_schema


async def to_code(config):
    varid = config[CONF_ID]
    metadata = get_mapping_metadata(varid.id)
    entries = {
        metadata.from_.conversion(key): await metadata.to_.convert_value(value)
        for key, value in config[CONF_ENTRIES].items()
    }
    value_type = metadata.to_.data_type
    # entries guaranteed to be non-empty here.
    value_0 = list(entries.values())[0]
    if isinstance(value_0, MockObj) and value_0.op != ".":
        value_type = value_type.operator("ptr")
    varid.type = mapping_class.template(
        metadata.from_.data_type,
        value_type,
    )
    var = MockObj(varid, ".")
    decl = VariableDeclarationExpression(varid.type, "", varid, static=True)
    add_global(decl)
    CORE.register_variable(varid, var)

    for key, value in entries.items():
        cg.add(var.set(key, value))
    if (default_value := config.get(CONF_DEFAULT_VALUE)) is not None:
        cg.add(var.set_default_value(await metadata.to_.convert_value(default_value)))
    return var



# ---------------------------------------------------------------------------
# Legacy top-level `online_image:` deprecation shim -- REMOVE this whole file
# after 2027.1.0.
#
# Online images are now a platform of the `image:` component (`platform:
# online_image`); the real schema, actions and codegen live in `image.py`. This
# module only keeps the deprecated top-level `online_image:` key working during
# the deprecation window: it reuses that schema/codegen and adds a one-shot
# deprecation warning (with a pasteable migrated `image:` block) at validation
# time. Deleting this file drops the top-level form entirely.
# ---------------------------------------------------------------------------

import esphome.components.image as espImage
import esphome.config_validation as cv

from .image import ONLINE_IMAGE_CONFIG_SCHEMA, setup_online_image

AUTO_LOAD = ["image", "runtime_image"]
DEPENDENCIES = ["display", "http_request"]
CODEOWNERS = ["@guillempages", "@clydebarrow"]
MULTI_CONF = True

DOMAIN = "online_image"

LEGACY_REMOVAL_VERSION = "2027.1.0"

_capture_legacy_entry, _warn_legacy_online_image = (
    espImage.legacy_platform_migration_warning(DOMAIN, DOMAIN, LEGACY_REMOVAL_VERSION)
)

CONFIG_SCHEMA = cv.All(_capture_legacy_entry, ONLINE_IMAGE_CONFIG_SCHEMA)

FINAL_VALIDATE_SCHEMA = _warn_legacy_online_image

to_code = setup_online_image



import esphome.config_validation as cv

CONFIG_SCHEMA = cv.invalid(
    "The I2S audio media player has been removed. "
    "Use the speaker media player component instead. "
    "See https://esphome.io/components/media_player/speaker.html for details."
)



import esphome.codegen as cg
from esphome.components import fan
import esphome.config_validation as cv

from .. import BEDJET_CLIENT_SCHEMA, bedjet_ns, register_bedjet_child

CODEOWNERS = ["@jhansche"]
DEPENDENCIES = ["bedjet"]

BedJetFan = bedjet_ns.class_("BedJetFan", fan.Fan, cg.PollingComponent)

CONFIG_SCHEMA = (
    fan.fan_schema(BedJetFan)
    .extend(cv.polling_component_schema("60s"))
    .extend(BEDJET_CLIENT_SCHEMA)
)


async def to_code(config):
    var = await fan.new_fan(config)
    await cg.register_component(var, config)
    await register_bedjet_child(var, config)



import datetime
import random

import esphome.codegen as cg
from esphome.components.zephyr import zephyr_add_prj_conf
import esphome.config_validation as cv
from esphome.const import (
    CONF_ID,
    CONF_MODEL,
    CONF_NAME,
    CONF_UNIT_OF_MEASUREMENT,
    __version__,
)
from esphome.core import CORE, CoroPriority, coroutine_with_priority
from esphome.cpp_generator import (
    AssignmentExpression,
    MockObj,
    VariableDeclarationExpression,
)
from esphome.types import ConfigType

from .const import (
    BACNET_UNIT_NO_UNITS,
    BACNET_UNITS,
    CONF_POWER_SOURCE,
    CONF_ROUTER,
    CONF_WIPE_ON_BOOT,
    KEY_ZIGBEE,
    POWER_SOURCE,
    AnalogAttrs,
    AnalogAttrsOutput,
    BinaryAttrs,
    ZigbeeComponent,
    zigbee_ns,
)
from .const_zephyr import (
    CONF_IEEE802154_VENDOR_OUI,
    CONF_SLEEPY,
    CONF_ZIGBEE_BINARY_SENSOR,
    CONF_ZIGBEE_ID,
    CONF_ZIGBEE_NUMBER,
    CONF_ZIGBEE_SENSOR,
    CONF_ZIGBEE_SWITCH,
    KEY_EP_NUMBER,
    ZB_ZCL_BASIC_ATTRS_EXT_T,
    ZB_ZCL_CLUSTER_ID_ANALOG_INPUT,
    ZB_ZCL_CLUSTER_ID_ANALOG_OUTPUT,
    ZB_ZCL_CLUSTER_ID_BASIC,
    ZB_ZCL_CLUSTER_ID_BINARY_INPUT,
    ZB_ZCL_CLUSTER_ID_BINARY_OUTPUT,
    ZB_ZCL_CLUSTER_ID_IDENTIFY,
    ZB_ZCL_IDENTIFY_ATTRS_T,
)

ZigbeeBinarySensor = zigbee_ns.class_("ZigbeeBinarySensor", cg.Component)
ZigbeeSensor = zigbee_ns.class_("ZigbeeSensor", cg.Component)
ZigbeeSwitch = zigbee_ns.class_("ZigbeeSwitch", cg.Component)
ZigbeeNumber = zigbee_ns.class_("ZigbeeNumber", cg.Component)

zephyr_binary_sensor = cv.Schema(
    {
        cv.OnlyWith(CONF_ZIGBEE_ID, ["nrf52", "zigbee"]): cv.use_id(ZigbeeComponent),
        cv.OnlyWith(CONF_ZIGBEE_BINARY_SENSOR, ["nrf52", "zigbee"]): cv.declare_id(
            ZigbeeBinarySensor
        ),
    }
)

zephyr_sensor = cv.Schema(
    {
        cv.OnlyWith(CONF_ZIGBEE_ID, ["nrf52", "zigbee"]): cv.use_id(ZigbeeComponent),
        cv.OnlyWith(CONF_ZIGBEE_SENSOR, ["nrf52", "zigbee"]): cv.declare_id(
            ZigbeeSensor
        ),
    }
)

zephyr_switch = cv.Schema(
    {
        cv.OnlyWith(CONF_ZIGBEE_ID, ["nrf52", "zigbee"]): cv.use_id(ZigbeeComponent),
        cv.OnlyWith(CONF_ZIGBEE_SWITCH, ["nrf52", "zigbee"]): cv.declare_id(
            ZigbeeSwitch
        ),
    }
)

zephyr_number = cv.Schema(
    {
        cv.OnlyWith(CONF_ZIGBEE_ID, ["nrf52", "zigbee"]): cv.use_id(ZigbeeComponent),
        cv.OnlyWith(CONF_ZIGBEE_NUMBER, ["nrf52", "zigbee"]): cv.declare_id(
            ZigbeeNumber
        ),
    }
)


async def zephyr_to_code(config: ConfigType) -> "MockObj":
    zephyr_add_prj_conf("ZIGBEE", True)
    zephyr_add_prj_conf("ZIGBEE_APP_UTILS", True)
    if config[CONF_ROUTER]:
        zephyr_add_prj_conf("ZIGBEE_ROLE_ROUTER", True)
    else:
        zephyr_add_prj_conf("ZIGBEE_ROLE_END_DEVICE", True)

    zephyr_add_prj_conf("ZIGBEE_CHANNEL_SELECTION_MODE_MULTI", True)

    zephyr_add_prj_conf("CRYPTO", True)

    zephyr_add_prj_conf("NET_IPV6", False)
    zephyr_add_prj_conf("NET_IP_ADDR_CHECK", False)
    zephyr_add_prj_conf("NET_UDP", False)

    # disable all extra to reduce power and save flash
    zephyr_add_prj_conf("ZIGBEE_HAVE_SERIAL", False)
    zephyr_add_prj_conf("ZBOSS_ERROR_PRINT_TO_LOG", False)
    zephyr_add_prj_conf("DK_LIBRARY", False)

    cg.add_build_flag("-Wl,--wrap=zb_zcl_put_reporting_info_from_req")

    if CONF_IEEE802154_VENDOR_OUI in config:
        zephyr_add_prj_conf("IEEE802154_VENDOR_OUI_ENABLE", True)
        random_number = config[CONF_IEEE802154_VENDOR_OUI]
        if random_number == "random":
            random_number = random.randint(0x000000, 0xFFFFFF)
        zephyr_add_prj_conf("IEEE802154_VENDOR_OUI", random_number)

    if config[CONF_WIPE_ON_BOOT]:
        if config[CONF_WIPE_ON_BOOT] == "once":
            cg.add_define(
                "USE_ZIGBEE_WIPE_ON_BOOT_MAGIC", random.randint(0x000001, 0xFFFFFF)
            )
        cg.add_define("USE_ZIGBEE_WIPE_ON_BOOT")

    # Generate attribute lists before any await that could yield (e.g., build_automation
    # waiting for variables from other components). If the hub's priority decays while
    # yielding, deferred entity jobs may add cluster list globals that reference these
    # attribute lists before they're declared.
    await _attr_to_code(config)

    var = cg.new_Pvariable(config[CONF_ID])

    await cg.register_component(var, config)

    CORE.add_job(_ctx_to_code, config)

    cg.add(var.set_sleepy(config[CONF_SLEEPY]))

    return var


async def _attr_to_code(config: ConfigType) -> None:
    # Create the basic attributes structure and attribute list
    basic_attrs = zigbee_new_variable("zigbee_basic_attrs", ZB_ZCL_BASIC_ATTRS_EXT_T)
    zigbee_new_attr_list(
        "zigbee_basic_attrib_list",
        "ZB_ZCL_DECLARE_BASIC_ATTRIB_LIST_EXT",
        zigbee_assign(basic_attrs.zcl_version, cg.RawExpression("ZB_ZCL_VERSION")),
        zigbee_assign(basic_attrs.app_version, 0),
        zigbee_assign(basic_attrs.stack_version, 0),
        zigbee_assign(basic_attrs.hw_version, 0),
        zigbee_set_string(basic_attrs.mf_name, "esphome"),
        zigbee_set_string(basic_attrs.model_id, config[CONF_MODEL]),
        zigbee_set_string(
            basic_attrs.date_code,
            # Local build time, matching the esp32 implementation
            # (App.get_build_time() in C++).
            datetime.datetime.now().astimezone().strftime("%Y%m%d %H%M%S"),
        ),
        zigbee_assign(
            basic_attrs.power_source,
            POWER_SOURCE[config[CONF_POWER_SOURCE]],
        ),
        zigbee_set_string(basic_attrs.location_id, ""),
        zigbee_assign(
            basic_attrs.ph_env, cg.RawExpression("ZB_ZCL_BASIC_ENV_UNSPECIFIED")
        ),
        zigbee_set_string(basic_attrs.sw_ver, __version__),
    )

    # Create the identify attributes structure and attribute list
    identify_attrs = zigbee_new_variable(
        "zigbee_identify_attrs", ZB_ZCL_IDENTIFY_ATTRS_T
    )
    zigbee_new_attr_list(
        "zigbee_identify_attrib_list",
        "ZB_ZCL_DECLARE_IDENTIFY_ATTRIB_LIST",
        zigbee_assign(
            identify_attrs.identify_time,
            cg.RawExpression("ZB_ZCL_IDENTIFY_IDENTIFY_TIME_DEFAULT_VALUE"),
        ),
    )


def zigbee_new_variable(name: str, type_: str) -> cg.MockObj:
    """Create a global variable with the given name and type."""
    decl = VariableDeclarationExpression(type_, "", name)
    CORE.add_global(decl)
    return MockObj(name, ".")


def zigbee_assign(target: cg.MockObj, expression: cg.RawExpression | int) -> str:
    """Assign an expression to a target and return a reference to it."""
    cg.add(AssignmentExpression("", "", target, expression))
    return f"&{target}"


def zigbee_set_string(target: cg.MockObj, value: str) -> str:
    """Set a ZCL string value and return the target name (arrays decay to pointers)."""
    # Zigbee supports only ASCII
    value = value.encode("ascii", "ignore").decode()
    cg.add(
        cg.RawExpression(
            f"ZB_ZCL_SET_STRING_VAL({target}, {cg.safe_exp(value)}, ZB_ZCL_STRING_CONST_SIZE({cg.safe_exp(value)}))"
        )
    )
    return str(target)


def zigbee_new_attr_list(name: str, macro: str, *args: str) -> str:
    """Create an attribute list using a ZBOSS macro and return the name."""
    obj = cg.RawExpression(f"{macro}({name}, {', '.join(args)})")
    CORE.add_global(obj)
    return name


class ZigbeeClusterDesc:
    """Represents a Zigbee cluster descriptor for code generation."""

    def __init__(self, cluster_id: str, attr_list_name: str | None = None) -> None:
        self._cluster_id = cluster_id
        self._attr_list_name = attr_list_name

    @property
    def cluster_id(self) -> str:
        return self._cluster_id

    @property
    def has_attrs(self) -> bool:
        return self._attr_list_name is not None

    def __str__(self) -> str:
        role = (
            "ZB_ZCL_CLUSTER_SERVER_ROLE"
            if self._attr_list_name
            else "ZB_ZCL_CLUSTER_CLIENT_ROLE"
        )
        if self._attr_list_name:
            attr_count = f"ZB_ZCL_ARRAY_SIZE({self._attr_list_name}, zb_zcl_attr_t)"
            return f"ZB_ZCL_CLUSTER_DESC({self._cluster_id}, {attr_count}, {self._attr_list_name}, {role}, ZB_ZCL_MANUF_CODE_INVALID)"
        return f"ZB_ZCL_CLUSTER_DESC({self._cluster_id}, 0, NULL, {role}, ZB_ZCL_MANUF_CODE_INVALID)"


def zigbee_new_cluster_list(
    name: str, clusters: list[ZigbeeClusterDesc]
) -> tuple[str, list[ZigbeeClusterDesc]]:
    """Create a cluster list array and return its name and the clusters."""
    # Always include basic and identify clusters first
    all_clusters = [
        ZigbeeClusterDesc(ZB_ZCL_CLUSTER_ID_BASIC, "zigbee_basic_attrib_list"),
        ZigbeeClusterDesc(ZB_ZCL_CLUSTER_ID_IDENTIFY, "zigbee_identify_attrib_list"),
    ]
    all_clusters.extend(clusters)

    cluster_strs = [str(c) for c in all_clusters]
    CORE.add_global(
        cg.RawExpression(
            f"zb_zcl_cluster_desc_t {name}[] = {{{', '.join(cluster_strs)}}}"
        )
    )
    return (name, all_clusters)


def zigbee_register_ep(
    ep_name: str,
    cluster_list_name: str,
    report_attr_count: int,
    clusters: list[ZigbeeClusterDesc],
    slot_index: int,
    app_device_id: str,
) -> None:
    """Register a Zigbee endpoint."""
    in_cluster_num = sum(1 for c in clusters if c.has_attrs)
    out_cluster_num = len(clusters) - in_cluster_num
    cluster_ids = [c.cluster_id for c in clusters]

    # Store endpoint name for device context generation
    CORE.data[KEY_ZIGBEE][KEY_EP_NUMBER][slot_index] = ep_name

    # Generate the endpoint declaration
    ep_id = slot_index + 1  # Endpoints are 1-indexed
    obj = cg.RawExpression(
        f"ESPHOME_ZB_HA_DECLARE_EP({ep_name}, {ep_id}, {cluster_list_name}, "
        f"{in_cluster_num}, {out_cluster_num}, {report_attr_count}, {app_device_id}, {', '.join(cluster_ids)})"
    )
    CORE.add_global(obj)


@coroutine_with_priority(CoroPriority.LATE)
async def _ctx_to_code(config: ConfigType) -> None:
    cg.add_define("ZIGBEE_ENDPOINTS_COUNT", len(CORE.data[KEY_ZIGBEE][KEY_EP_NUMBER]))
    cg.add_global(
        cg.RawExpression(
            f"ZBOSS_DECLARE_DEVICE_CTX_EP_VA(zb_device_ctx, &{', &'.join(CORE.data[KEY_ZIGBEE][KEY_EP_NUMBER])})"
        )
    )
    cg.add(cg.RawExpression("ZB_AF_REGISTER_DEVICE_CTX(&zb_device_ctx)"))


async def zephyr_setup_binary_sensor(entity: cg.MockObj, config: ConfigType) -> None:
    CORE.add_job(_add_binary_sensor, entity, config)


async def zephyr_setup_sensor(entity: cg.MockObj, config: ConfigType) -> None:
    CORE.add_job(_add_sensor, entity, config)


async def zephyr_setup_switch(entity: cg.MockObj, config: ConfigType) -> None:
    CORE.add_job(_add_switch, entity, config)


async def zephyr_setup_number(
    entity: cg.MockObj,
    config: ConfigType,
    min_value: float,
    max_value: float,
    step: float,
) -> None:
    CORE.add_job(_add_number, entity, config, min_value, max_value, step)


def get_slot_index() -> int:
    """Find the next available endpoint slot."""
    slot = next(
        (i for i, v in enumerate(CORE.data[KEY_ZIGBEE][KEY_EP_NUMBER]) if v == ""), None
    )
    if slot is None:
        raise cv.Invalid(
            f"No available Zigbee endpoint slots ({len(CORE.data[KEY_ZIGBEE][KEY_EP_NUMBER])} in use)"
        )
    return slot


async def _add_zigbee_ep(
    entity: cg.MockObj,
    config: ConfigType,
    component_key,
    attrs_type,
    zcl_macro: str,
    cluster_id: str,
    app_device_id: str,
    extra_field_values: dict[str, int] | None = None,
) -> None:
    slot_index = get_slot_index()

    prefix = f"zigbee_ep{slot_index + 1}"
    attrs_name = f"{prefix}_attrs"
    attr_list_name = f"{prefix}_attrib_list"
    cluster_list_name = f"{prefix}_cluster_list"
    ep_name = f"{prefix}_ep"

    # Create attribute struct
    attrs = zigbee_new_variable(attrs_name, attrs_type)

    # Build attribute list args
    attr_args = [
        zigbee_assign(attrs.out_of_service, 0),
        zigbee_assign(attrs.present_value, 0),
        zigbee_assign(attrs.status_flags, 0),
    ]
    # Add extra field assignments (e.g., engineering_units for sensors)
    if extra_field_values:
        for field_name, value in extra_field_values.items():
            attr_args.append(zigbee_assign(getattr(attrs, field_name), value))
    attr_args.append(zigbee_set_string(attrs.description, config[CONF_NAME]))

    # Create attribute list
    attr_list = zigbee_new_attr_list(attr_list_name, zcl_macro, *attr_args)

    # Create cluster list and register endpoint
    cluster_list_name, clusters = zigbee_new_cluster_list(
        cluster_list_name,
        [ZigbeeClusterDesc(cluster_id, attr_list)],
    )
    zigbee_register_ep(
        ep_name, cluster_list_name, 2, clusters, slot_index, app_device_id
    )

    # Create ESPHome component
    var = cg.new_Pvariable(config[component_key], entity)
    await cg.register_component(var, {})

    cg.add(var.set_endpoint(slot_index + 1))
    cg.add(var.set_cluster_attributes(attrs))

    hub = await cg.get_variable(config[CONF_ZIGBEE_ID])
    cg.add(var.set_parent(hub))


async def _add_binary_sensor(entity: cg.MockObj, config: ConfigType) -> None:
    await _add_zigbee_ep(
        entity,
        config,
        CONF_ZIGBEE_BINARY_SENSOR,
        BinaryAttrs,
        "ESPHOME_ZB_ZCL_DECLARE_BINARY_INPUT_ATTRIB_LIST",
        ZB_ZCL_CLUSTER_ID_BINARY_INPUT,
        "ZB_HA_SIMPLE_SENSOR_DEVICE_ID",
    )


async def _add_sensor(entity: cg.MockObj, config: ConfigType) -> None:
    # Get BACnet engineering unit from unit_of_measurement
    unit = config.get(CONF_UNIT_OF_MEASUREMENT, "")
    bacnet_unit = BACNET_UNITS.get(unit, BACNET_UNIT_NO_UNITS)

    await _add_zigbee_ep(
        entity,
        config,
        CONF_ZIGBEE_SENSOR,
        AnalogAttrs,
        "ESPHOME_ZB_ZCL_DECLARE_ANALOG_INPUT_ATTRIB_LIST",
        ZB_ZCL_CLUSTER_ID_ANALOG_INPUT,
        "ZB_HA_CUSTOM_ATTR_DEVICE_ID",
        extra_field_values={"engineering_units": bacnet_unit},
    )


async def _add_switch(entity: cg.MockObj, config: ConfigType) -> None:
    await _add_zigbee_ep(
        entity,
        config,
        CONF_ZIGBEE_SWITCH,
        BinaryAttrs,
        "ESPHOME_ZB_ZCL_DECLARE_BINARY_OUTPUT_ATTRIB_LIST",
        ZB_ZCL_CLUSTER_ID_BINARY_OUTPUT,
        "ZB_HA_CUSTOM_ATTR_DEVICE_ID",
    )


async def _add_number(
    entity: cg.MockObj,
    config: ConfigType,
    min_value: float,
    max_value: float,
    step: float,
) -> None:
    # Get BACnet engineering unit from unit_of_measurement
    unit = config.get(CONF_UNIT_OF_MEASUREMENT, "")
    bacnet_unit = BACNET_UNITS.get(unit, BACNET_UNIT_NO_UNITS)

    await _add_zigbee_ep(
        entity,
        config,
        CONF_ZIGBEE_NUMBER,
        AnalogAttrsOutput,
        "ESPHOME_ZB_ZCL_DECLARE_ANALOG_OUTPUT_ATTRIB_LIST",
        ZB_ZCL_CLUSTER_ID_ANALOG_OUTPUT,
        "ZB_HA_CUSTOM_ATTR_DEVICE_ID",
        extra_field_values={
            "max_present_value": max_value,
            "min_present_value": min_value,
            "resolution": step,
            "engineering_units": bacnet_unit,
        },
    )



import esphome.codegen as cg
from esphome.components import i2c
from esphome.components.motion import MotionComponent

CODEOWNERS = ["@clydebarrow"]

CONF_BMI270_ID = "bmi270_id"
#  C++ namespace / class
bmi270_ns = cg.esphome_ns.namespace("bmi270")
BMI270Component = bmi270_ns.class_("BMI270Component", MotionComponent, i2c.I2CDevice)



import esphome.codegen as cg
from esphome.components import number
import esphome.config_validation as cv
from esphome.const import CONF_ON_RELEASE, CONF_RESTORE_VALUE
from esphome.cpp_generator import MockObj

from ..defines import (
    CONF_ANIMATED,
    CONF_TRIGGER,
    CONF_UPDATE_ON_RELEASE,
    CONF_WIDGET,
    LOGGER,
)
from ..lv_validation import animated
from ..lvcode import (
    EVENT_ARG,
    UPDATE_EVENT,
    LambdaContext,
    ReturnStatement,
    lv_obj,
    lvgl_static,
)
from ..schemas import TRIGGER_EVENT_MAP, VALUE_TRIGGER_SCHEMA
from ..types import LvNumber, lvgl_ns
from ..widgets import get_widgets, wait_for_widgets

LVGLNumber = lvgl_ns.class_("LVGLNumber", number.Number, cg.Component)

CONFIG_SCHEMA = number.number_schema(LVGLNumber).extend(
    {
        cv.Required(CONF_WIDGET): cv.use_id(LvNumber),
        **VALUE_TRIGGER_SCHEMA,
        cv.Optional(CONF_ANIMATED, default=True): animated,
        cv.Optional(CONF_UPDATE_ON_RELEASE): cv.boolean,
        cv.Optional(CONF_RESTORE_VALUE, default=False): cv.boolean,
    }
)


async def to_code(config):
    trigger = config[CONF_TRIGGER]
    if CONF_UPDATE_ON_RELEASE in config:
        LOGGER.warning(
            "Option 'update_on_release' is deprecated and will be removed in 2026.11.0 - use 'trigger: on_release' instead"
        )
        if config[CONF_UPDATE_ON_RELEASE]:
            trigger = CONF_ON_RELEASE
    widget = await get_widgets(config, CONF_WIDGET)
    widget = widget[0]
    await wait_for_widgets()
    async with LambdaContext([], return_type=cg.float_) as value:
        value.add(ReturnStatement(widget.get_value()))
    async with LambdaContext([(cg.float_, "v")]) as control:
        await widget.set_property(
            "value", MockObj("v") * MockObj(widget.get_scale()), config[CONF_ANIMATED]
        )
        lv_obj.send_event(widget.obj, UPDATE_EVENT, cg.nullptr)
    var = await number.new_number(
        config,
        await control.get_lambda(),
        await value.get_lambda(),
        config[CONF_RESTORE_VALUE],
        max_value=await widget.type.get_max(widget.config),
        min_value=await widget.type.get_min(widget.config),
        step=widget.type.get_step(widget.config),
    )
    async with LambdaContext(EVENT_ARG) as event:
        event.add(var.on_value())
    await cg.register_component(var, config)
    cg.add(
        lvgl_static.add_event_cb(
            widget.obj,
            await event.get_lambda(),
            *TRIGGER_EVENT_MAP[trigger],
        )
    )



"""ESP-IDF direct build generator for ESPHome."""

import json
from pathlib import Path

from esphome.components.esp32 import get_esp32_variant, idf_version
import esphome.config_validation as cv
from esphome.core import CORE
from esphome.framework_helpers import (
    get_project_compile_flags,
    get_project_cxx_compile_flags,
    get_project_link_flags,
)
from esphome.helpers import mkdir_p, write_file_if_changed

# Replaces the IDF default C++ standard (-std=gnu++2b appended to
# CXX_COMPILE_OPTIONS by project.cmake's __build_init) with the one set via
# cg.set_cpp_standard(). Emitted between include(project.cmake) and project(),
# i.e. after IDF appends its default and before the options are consumed, and
# applies project-wide like PlatformIO build_unflags.
CPP_STANDARD_TEMPLATE = """\
idf_build_get_property(esphome_cxx_compile_options CXX_COMPILE_OPTIONS)
list(FILTER esphome_cxx_compile_options EXCLUDE REGEX "^-std=")
list(APPEND esphome_cxx_compile_options "-std={standard}")
idf_build_set_property(CXX_COMPILE_OPTIONS "${{esphome_cxx_compile_options}}")"""


def get_available_components() -> list[str] | None:
    """Get list of built-in ESP-IDF components from project_description.json.

    Excludes ``src``, IDF-managed components (``managed_components/``), and
    converted PIO libs (``pio_components/``). Returns ``None`` if the build
    dir or ``project_description.json`` isn't ready yet.
    """
    if CORE.build_path is None:
        return None
    project_desc = Path(CORE.build_path) / "build" / "project_description.json"
    if not project_desc.exists():
        return None

    try:
        with project_desc.open(encoding="utf-8") as f:
            data = json.load(f)

        component_info = data.get("build_component_info", {})

        result = []
        for name, info in component_info.items():
            # Exclude our own src component
            if name == "src":
                continue

            # Exclude IDF-managed and converted-PIO components (external).
            comp_dir = info.get("dir", "")
            if "managed_components" in comp_dir or "pio_components" in comp_dir:
                continue

            result.append(name)

        return result
    except (json.JSONDecodeError, OSError):
        return None


def has_discovered_components() -> bool:
    """Check if we have discovered components from a previous configure."""
    return get_available_components() is not None


def get_project_cmakelists(minimal: bool = False) -> str:
    """Generate the top-level CMakeLists.txt for ESP-IDF project.

    When ``minimal`` is true, omit ``ESPHOME_PROJECT_BUILTIN_COMPONENTS``
    since ``project_description.json`` may be stale on the first write.
    """
    # Get IDF target from ESP32 variant (e.g., ESP32S3 -> esp32s3)
    variant = get_esp32_variant()
    idf_target = variant.lower().replace("-", "")

    # esp_idf_size 2.x (bundled with IDF >=6.0) made NG the default and
    # removed the --ng flag; on 1.x (IDF 5.5) --ng is required to get
    # --format=raw because the legacy mode doesn't support it.
    size_ng_flag = "--ng" if idf_version() < cv.Version(6, 0, 0) else ""

    # Project-wide compile options: -D defines and -W warning flags (skip
    # -Wl, linker flags — those go on the src component via
    # target_link_options below). Emitted via idf_build_set_property so the
    # flags propagate to every IDF component (including managed ones like
    # esphome__micro-mp3) rather than just src/. Required so suppressions
    # like ``-Wno-error=maybe-uninitialized`` actually silence warnings in
    # third-party components we don't author.
    project_compile_opts = get_project_compile_flags()
    extra_compile_options = "\n".join(
        f'idf_build_set_property(COMPILE_OPTIONS "{flag}" APPEND)'
        for flag in project_compile_opts
    )

    # Flags registered via cg.add_cxx_build_flag() go on CXX_COMPILE_OPTIONS
    # (not COMPILE_OPTIONS) because GCC warns when a C++-only flag such as
    # -Wno-volatile is passed on a C compile.
    cxx_compile_options = "\n".join(
        f'idf_build_set_property(CXX_COMPILE_OPTIONS "{flag}" APPEND)'
        for flag in get_project_cxx_compile_flags()
    )

    cpp_standard_options = (
        CPP_STANDARD_TEMPLATE.format(standard=CORE.cpp_standard)
        if CORE.cpp_standard
        else ""
    )

    # Per-project list exposed as a CMake variable so converted PIO libs
    # can reference ${ESPHOME_PROJECT_MANAGED_COMPONENTS} without baking
    # project-specific names into their cached CMakeLists.
    #
    # Emit via idf_build_set_property (not plain set()) so the value is
    # serialised into build_properties.temp.cmake and visible to IDF's
    # early requirements-expansion pass (component_get_requirements.cmake
    # runs as a separate CMake script invocation that doesn't load the
    # project's top-level CMakeLists; without this, ${ESPHOME_PROJECT_
    # MANAGED_COMPONENTS} in a converted-lib REQUIRES expands to empty).
    from esphome.components.esp32 import get_managed_component_require_names

    managed_components_property = "\n".join(
        f"idf_build_set_property(ESPHOME_PROJECT_MANAGED_COMPONENTS {name} APPEND)"
        for name in get_managed_component_require_names()
    )

    # Built-in IDF components exposed via our own property (not IDF's
    # __COMPONENT_REQUIRES_COMMON, which would append them to every
    # component's REQUIRES including real IDF components). Referenced by
    # src/CMakeLists and by each converted PIO lib's CMakeLists. Skipped
    # on minimal writes because project_description.json may be stale.
    builtin_components_property = (
        ""
        if minimal
        else "\n".join(
            f"idf_build_set_property(ESPHOME_PROJECT_BUILTIN_COMPONENTS {name} APPEND)"
            for name in sorted(get_available_components() or [])
        )
    )

    return f"""\
# Auto-generated by ESPHome
cmake_minimum_required(VERSION 3.16)

# On Windows, Ninja can fail with:
#   "CreateProcess: The parameter is incorrect (is the command line too long?)"
# when compiler/linker command lines exceed the OS length limit.
#
# The following settings force CMake/Ninja to use *response files* (@file.rsp)
# to pass long lists of includes, objects, and other arguments indirectly,
# avoiding command-line length limits and fixing the build failure.
#
# This is especially useful for large ESP-IDF / ESPHome projects with many
# source files or include directories.
set(CMAKE_C_USE_RESPONSE_FILE_FOR_INCLUDES 1)
set(CMAKE_CXX_USE_RESPONSE_FILE_FOR_INCLUDES 1)
set(CMAKE_C_USE_RESPONSE_FILE_FOR_OBJECTS 1)
set(CMAKE_CXX_USE_RESPONSE_FILE_FOR_OBJECTS 1)
set(CMAKE_NINJA_FORCE_RESPONSE_FILE 1)

set(IDF_TARGET {idf_target})
set(EXTRA_COMPONENT_DIRS ${{CMAKE_SOURCE_DIR}}/src)

include($ENV{{IDF_PATH}}/tools/cmake/project.cmake)

{cpp_standard_options}

{cxx_compile_options}

{extra_compile_options}

{managed_components_property}

{builtin_components_property}

project({CORE.name})

# Emit raw JSON size data for ESPHome to read post-build.
add_custom_command(
    TARGET ${{CMAKE_PROJECT_NAME}}.elf POST_BUILD
    COMMAND ${{PYTHON}} -m esp_idf_size {size_ng_flag} --format=raw
            -o ${{CMAKE_BINARY_DIR}}/esp_idf_size.json
            ${{CMAKE_PROJECT_NAME}}.map
    WORKING_DIRECTORY ${{CMAKE_BINARY_DIR}}
    VERBATIM
)
"""


def get_component_cmakelists() -> str:
    """Generate the main component CMakeLists.txt.

    REQUIRES pulls in the discovered built-in IDF components via the
    project-level variables set in the top-level CMakeLists.
    """
    # Extract linker options (-Wl, flags). Compile flags (-D, -W) are
    # emitted project-wide via idf_build_set_property in
    # get_project_cmakelists so they reach every component, not just src/.
    link_opts = get_project_link_flags()
    link_opts_str = "\n    ".join(link_opts) if link_opts else ""

    return f"""\
# Auto-generated by ESPHome
# CONFIGURE_DEPENDS asks CMake to re-check the glob each build so test
# runs that reuse the build dir don't compile stale source paths. It's
# invalid in script mode (cmake -P), which is how IDF's
# component_get_requirements.cmake includes us, so skip it there.
if(CMAKE_SCRIPT_MODE_FILE)
    file(GLOB_RECURSE app_sources
        "${{CMAKE_CURRENT_SOURCE_DIR}}/*.cpp"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/*.cc"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/*.cxx"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/*.c++"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/*.c"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/esphome/*.cpp"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/esphome/*.cc"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/esphome/*.cxx"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/esphome/*.c++"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/esphome/*.c"
    )
else()
    file(GLOB_RECURSE app_sources CONFIGURE_DEPENDS
        "${{CMAKE_CURRENT_SOURCE_DIR}}/*.cpp"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/*.cc"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/*.cxx"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/*.c++"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/*.c"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/esphome/*.cpp"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/esphome/*.cc"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/esphome/*.cxx"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/esphome/*.c++"
        "${{CMAKE_CURRENT_SOURCE_DIR}}/esphome/*.c"
    )
endif()

idf_component_register(
    SRCS ${{app_sources}}
    INCLUDE_DIRS "." "esphome"
    REQUIRES ${{ESPHOME_PROJECT_BUILTIN_COMPONENTS}}
)

# ESPHome linker options
target_link_options(${{COMPONENT_LIB}} PUBLIC
    {link_opts_str}
)
"""


def write_project(minimal: bool = False) -> None:
    """Write ESP-IDF project files."""
    mkdir_p(CORE.build_path)
    mkdir_p(CORE.relative_src_path())

    # Write top-level CMakeLists.txt
    write_file_if_changed(
        CORE.relative_build_path("CMakeLists.txt"),
        get_project_cmakelists(minimal=minimal),
    )

    # Write component CMakeLists.txt in src/
    write_file_if_changed(
        CORE.relative_src_path("CMakeLists.txt"),
        get_component_cmakelists(),
    )



from esphome import pins
import esphome.codegen as cg
from esphome.components import i2c, sensor
import esphome.config_validation as cv
from esphome.const import (
    CONF_FILTER,
    CONF_GAIN,
    CONF_ID,
    CONF_OVERSAMPLING,
    CONF_RESOLUTION,
    CONF_TEMPERATURE,
    CONF_TEMPERATURE_COMPENSATION,
    DEVICE_CLASS_TEMPERATURE,
    ICON_MAGNET,
    ICON_THERMOMETER,
    STATE_CLASS_MEASUREMENT,
    UNIT_CELSIUS,
    UNIT_MICROTESLA,
)

CODEOWNERS = ["@functionpointer"]
DEPENDENCIES = ["i2c"]

mlx90393_ns = cg.esphome_ns.namespace("mlx90393")

MLX90393Component = mlx90393_ns.class_(
    "MLX90393Cls", cg.PollingComponent, i2c.I2CDevice
)

GAIN = {
    "1X": 0,
    "1_25X": 1,
    "1_67X": 2,
    "2X": 3,
    "2_5X": 4,
    "3X": 5,
    "3_75X": 6,
    "5X": 7,
}

RESOLUTION = {
    "DIV_8": 3,
    "DIV_4": 2,
    "DIV_2": 1,
    "DIV_1": 0,
}

CONF_X_AXIS = "x_axis"
CONF_Y_AXIS = "y_axis"
CONF_Z_AXIS = "z_axis"
CONF_DRDY_PIN = "drdy_pin"
CONF_HALLCONF = "hallconf"


def _validate(config):
    if config[CONF_TEMPERATURE_COMPENSATION]:
        for axis in [CONF_X_AXIS, CONF_Y_AXIS, CONF_Z_AXIS]:
            if axis not in config:
                continue
            if (res := config[axis][CONF_RESOLUTION]) in [
                "DIV_8",
                "DIV_4",
            ]:
                raise cv.Invalid(
                    f"{axis}: {CONF_RESOLUTION} cannot be {res} with {CONF_TEMPERATURE_COMPENSATION} enabled"
                )
    if config[CONF_HALLCONF] == 0xC and (
        config[CONF_OVERSAMPLING],
        config[CONF_FILTER],
    ) in [(0, 0), (1, 0), (0, 1)]:
        raise cv.Invalid(
            f"{CONF_OVERSAMPLING}=={config[CONF_OVERSAMPLING]} and {CONF_FILTER}=={config[CONF_FILTER]} not allowed with {CONF_HALLCONF}=={config[CONF_HALLCONF]:#02x}"
        )
    return config


def mlx90393_axis_schema():
    return sensor.sensor_schema(
        unit_of_measurement=UNIT_MICROTESLA,
        accuracy_decimals=0,
        icon=ICON_MAGNET,
        state_class=STATE_CLASS_MEASUREMENT,
    ).extend(
        cv.Schema(
            {
                cv.Optional(CONF_RESOLUTION, default="DIV_4"): cv.enum(
                    RESOLUTION, upper=True, space="_"
                )
            }
        )
    )


CONFIG_SCHEMA = cv.All(
    cv.Schema(
        {
            cv.GenerateID(): cv.declare_id(MLX90393Component),
            cv.Optional(CONF_GAIN, default="1X"): cv.enum(GAIN, upper=True, space="_"),
            cv.Optional(CONF_DRDY_PIN): pins.gpio_input_pin_schema,
            cv.Optional(CONF_OVERSAMPLING, default=0): cv.int_range(min=0, max=3),
            cv.Optional(CONF_FILTER, default=6): cv.int_range(min=0, max=7),
            cv.Optional(CONF_X_AXIS): mlx90393_axis_schema(),
            cv.Optional(CONF_Y_AXIS): mlx90393_axis_schema(),
            cv.Optional(CONF_Z_AXIS): mlx90393_axis_schema(),
            cv.Optional(CONF_TEMPERATURE_COMPENSATION, default=False): bool,
            cv.Optional(CONF_HALLCONF, default=0xC): cv.one_of(0xC, 0x0),
            cv.Optional(CONF_TEMPERATURE): sensor.sensor_schema(
                unit_of_measurement=UNIT_CELSIUS,
                accuracy_decimals=1,
                icon=ICON_THERMOMETER,
                device_class=DEVICE_CLASS_TEMPERATURE,
                state_class=STATE_CLASS_MEASUREMENT,
            ).extend(
                cv.Schema(
                    {
                        cv.Optional(CONF_OVERSAMPLING, default=0): cv.int_range(
                            min=0, max=3
                        ),
                    }
                )
            ),
        },
    )
    .extend(cv.polling_component_schema("60s"))
    .extend(i2c.i2c_device_schema(0x0C)),
    _validate,
)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)
    await i2c.register_i2c_device(var, config)

    cg.add(var.set_gain(GAIN[config[CONF_GAIN]]))
    cg.add(var.set_oversampling(config[CONF_OVERSAMPLING]))
    cg.add(var.set_filter(config[CONF_FILTER]))
    cg.add(var.set_temperature_compensation(config[CONF_TEMPERATURE_COMPENSATION]))
    cg.add(var.set_hallconf(config[CONF_HALLCONF]))

    if CONF_X_AXIS in config:
        sens = await sensor.new_sensor(config[CONF_X_AXIS])
        cg.add(var.set_x_sensor(sens))
        cg.add(var.set_resolution(0, RESOLUTION[config[CONF_X_AXIS][CONF_RESOLUTION]]))
    if CONF_Y_AXIS in config:
        sens = await sensor.new_sensor(config[CONF_Y_AXIS])
        cg.add(var.set_y_sensor(sens))
        cg.add(var.set_resolution(1, RESOLUTION[config[CONF_Y_AXIS][CONF_RESOLUTION]]))
    if CONF_Z_AXIS in config:
        sens = await sensor.new_sensor(config[CONF_Z_AXIS])
        cg.add(var.set_z_sensor(sens))
        cg.add(var.set_resolution(2, RESOLUTION[config[CONF_Z_AXIS][CONF_RESOLUTION]]))
    if CONF_TEMPERATURE in config:
        sens = await sensor.new_sensor(config[CONF_TEMPERATURE])
        cg.add(var.set_t_sensor(sens))
        cg.add(var.set_t_oversampling(config[CONF_TEMPERATURE][CONF_OVERSAMPLING]))
    if CONF_DRDY_PIN in config:
        pin = await cg.gpio_pin_expression(config[CONF_DRDY_PIN])
        cg.add(var.set_drdy_gpio(pin))

    cg.add_library("functionpointer/arduino-MLX90393", "1.0.2")



from esphome import pins
import esphome.codegen as cg
import esphome.config_validation as cv
from esphome.const import CONF_DELAY, CONF_ID

AUTO_LOAD = ["sensor", "voltage_sampler"]
CODEOWNERS = ["@asoehlke"]
MULTI_CONF = True

cd74hc4067_ns = cg.esphome_ns.namespace("cd74hc4067")

CD74HC4067Component = cd74hc4067_ns.class_("CD74HC4067Component", cg.Component)

CONF_PIN_S0 = "pin_s0"
CONF_PIN_S1 = "pin_s1"
CONF_PIN_S2 = "pin_s2"
CONF_PIN_S3 = "pin_s3"

DEFAULT_DELAY = "2ms"


CONFIG_SCHEMA = cv.Schema(
    {
        cv.GenerateID(): cv.declare_id(CD74HC4067Component),
        cv.Required(CONF_PIN_S0): pins.gpio_output_pin_schema,
        cv.Required(CONF_PIN_S1): pins.gpio_output_pin_schema,
        cv.Required(CONF_PIN_S2): pins.gpio_output_pin_schema,
        cv.Required(CONF_PIN_S3): pins.gpio_output_pin_schema,
        cv.Optional(
            CONF_DELAY, default=DEFAULT_DELAY
        ): cv.positive_time_period_milliseconds,
    }
).extend(cv.COMPONENT_SCHEMA)


async def to_code(config):
    var = cg.new_Pvariable(config[CONF_ID])
    await cg.register_component(var, config)

    pin_s0 = await cg.gpio_pin_expression(config[CONF_PIN_S0])
    cg.add(var.set_pin_s0(pin_s0))
    pin_s1 = await cg.gpio_pin_expression(config[CONF_PIN_S1])
    cg.add(var.set_pin_s1(pin_s1))
    pin_s2 = await cg.gpio_pin_expression(config[CONF_PIN_S2])
    cg.add(var.set_pin_s2(pin_s2))
    pin_s3 = await cg.gpio_pin_expression(config[CONF_PIN_S3])
    cg.add(var.set_pin_s3(pin_s3))

    cg.add(var.set_switch_delay(config[CONF_DELAY]))



"""Zephyr backend for the shared PlatformIO library converter.

For each PlatformIO library added via ``cg.add_library()``, emit a Zephyr
external module (``zephyr/module.yml`` + ``zephyr/CMakeLists.txt`` built with the
``zephyr_library*`` API) into the shared ``pio_components`` cache. The caller
wires the resulting module directories into the build via
``EXTRA_ZEPHYR_MODULES``; Zephyr then compiles each module and links it into the
final image.

Only framework-agnostic libraries (plain C/C++ that doesn't depend on the Arduino
API) will actually compile under Zephyr — this converter shares the
fetch/parse/cache plumbing, not API compatibility.
"""

from pathlib import Path

from esphome import yaml_util
from esphome.core import EsphomeError, Library
from esphome.helpers import write_file_if_changed
from esphome.platformio.library import (
    DEFAULT_BUILD_FLAGS,
    DEFAULT_BUILD_INCLUDE_DIR,
    DEFAULT_BUILD_SRC_FILTER,
    SRC_FILE_EXTENSIONS,
    ConvertedLibrary,
    LibraryBackend,
    PathType,
    collect_filtered_files,
    convert_libraries,
    ensure_list,
    split_list_by_condition,
)

# Zephyr libraries declare frameworks rarely and the PIO ``platforms`` token for
# nRF is seldom present, so the platform check is disabled (None) and only the
# framework mismatch warning fires.
ZEPHYR_FRAMEWORK = "zephyr"


def _escape(p: PathType) -> str:
    # In CMakeLists.txt, backslashes need to be escaped (mirrors the ESP-IDF
    # backend's escape_entry). Doubling -- rather than rewriting '\' -> '/' --
    # preserves content, so it's safe for arbitrary build flags (e.g. a -D value
    # containing a backslash) as well as Windows paths.
    return f'"{str(p)}"'.replace("\\", "\\\\")


def generate_module_yml(component: ConvertedLibrary) -> str:
    """Render the ``zephyr/module.yml`` manifest for a converted library."""
    return yaml_util.dump(
        {
            "name": component.get_require_name(),
            "build": {"cmake": "zephyr"},
        }
    )


def generate_cmakelists_txt(component: ConvertedLibrary) -> str:
    """Render the ``zephyr/CMakeLists.txt`` that builds a converted library.

    Sources/includes are emitted as absolute paths since the CMakeLists lives in
    the library's ``zephyr/`` subdir while its sources sit alongside it. Include
    dirs are published globally so the app (and sibling libraries) can include the
    library's headers, mirroring ESP-IDF's public ``INCLUDE_DIRS``.
    """
    build = component.data.get("build", {})

    build_src_dir = build.get("srcDir")
    if not build_src_dir:
        for d in ["src", "Src", "."]:
            if (component.path / Path(d)).is_dir():
                build_src_dir = d
                break

    build_include_dir = build.get("includeDir", DEFAULT_BUILD_INCLUDE_DIR)
    build_src_filter = ensure_list(build.get("srcFilter", DEFAULT_BUILD_SRC_FILTER))
    build_flags = ensure_list(build.get("flags", DEFAULT_BUILD_FLAGS))

    src_files = collect_filtered_files(
        component.path / Path(build_src_dir), build_src_filter
    )
    src_files = sorted(
        str(Path(p).resolve())
        for p in src_files
        if Path(p).suffix in SRC_FILE_EXTENSIONS
    )

    include_dir_flags, build_flags = split_list_by_condition(
        build_flags, lambda a: a[2:].strip() if a.startswith("-I") else None
    )
    link_directories, build_flags = split_list_by_condition(
        build_flags, lambda a: a[2:].strip() if a.startswith("-L") else None
    )
    link_libraries, build_flags = split_list_by_condition(
        build_flags, lambda a: a[2:].strip() if a.startswith("-l") else None
    )

    include_dirs = [build_include_dir, build_src_dir, *include_dir_flags]
    include_dirs = [
        str((component.path / Path(d)).resolve())
        for d in include_dirs
        if (component.path / Path(d)).is_dir()
    ]

    lines = [f"zephyr_library_named({component.get_require_name()})"]
    if src_files:
        lines += [
            "zephyr_library_sources(",
            *[f"  {_escape(p)}" for p in src_files],
            ")",
        ]
    if include_dirs:
        lines += [
            "zephyr_include_directories(",
            *[f"  {_escape(p)}" for p in include_dirs],
            ")",
        ]
    if build_flags:
        lines += [
            "zephyr_library_compile_options(",
            *[f"  {_escape(f)}" for f in build_flags],
            ")",
        ]
    # Best-effort link wiring; most Zephyr-portable libraries don't need it.
    link_flags = [f"-L{d}" for d in link_directories] + [
        f"-l{lib}" for lib in link_libraries
    ]
    if link_flags:
        lines += [
            "zephyr_link_libraries(",
            *[f"  {_escape(f)}" for f in link_flags],
            ")",
        ]

    return "\n".join(lines) + "\n"


def _emit_zephyr_module(component: ConvertedLibrary) -> None:
    zephyr_dir = component.path / "zephyr"
    write_file_if_changed(zephyr_dir / "module.yml", generate_module_yml(component))
    write_file_if_changed(
        zephyr_dir / "CMakeLists.txt", generate_cmakelists_txt(component)
    )


def generate_zephyr_modules(libraries: list[Library]) -> list[Path]:
    """Convert ``libraries`` to Zephyr modules and return all module directories.

    The returned list includes transitive dependencies (each converted library is
    its own module). Every directory should be added to ``EXTRA_ZEPHYR_MODULES``;
    Zephyr links all module libraries into the image, so cross-library symbols
    resolve without explicit dependency declarations.

    Raises ``EsphomeError`` if two libraries resolve to the same Zephyr module
    name -- each module's CMakeLists calls ``zephyr_library_named(<name>)``, so a
    duplicate would otherwise fail the build with a CMake "target already exists".
    The converter already warns when a library is referenced under inconsistent
    specs (bare ``name`` vs ``owner/name``, git vs registry); this turns that into
    an actionable error at the Zephyr boundary where it is fatal.
    """
    module_dirs: list[Path] = []
    by_name: dict[str, Path] = {}

    def emit(component: ConvertedLibrary) -> None:
        name = component.get_require_name()
        if name in by_name:
            raise EsphomeError(
                f"Two libraries resolve to the same Zephyr module '{name}' "
                f"({by_name[name]} and {component.path}). Reference the library "
                f"consistently (e.g. always as 'owner/name') so it resolves once."
            )
        by_name[name] = component.path
        _emit_zephyr_module(component)
        module_dirs.append(component.path)

    backend = LibraryBackend(
        platform=None, framework=ZEPHYR_FRAMEWORK, emit=emit, cache_key="zephyr"
    )
    convert_libraries(libraries, backend)
    return module_dirs



import esphome.codegen as cg
from esphome.components import switch
import esphome.config_validation as cv
from esphome.const import CONF_TYPE, ENTITY_CATEGORY_CONFIG
from esphome.cpp_generator import MockObjClass

from .. import CONF_DFROBOT_SEN0395_ID, DfrobotSen0395Component

DEPENDENCIES = ["dfrobot_sen0395"]

dfrobot_sen0395_ns = cg.esphome_ns.namespace("dfrobot_sen0395")
DfrobotSen0395Switch = dfrobot_sen0395_ns.class_(
    "DfrobotSen0395Switch",
    switch.Switch,
    cg.Component,
    cg.Parented.template(DfrobotSen0395Component),
)

Sen0395PowerSwitch = dfrobot_sen0395_ns.class_(
    "Sen0395PowerSwitch", DfrobotSen0395Switch
)
Sen0395LedSwitch = dfrobot_sen0395_ns.class_("Sen0395LedSwitch", DfrobotSen0395Switch)
Sen0395UartPresenceSwitch = dfrobot_sen0395_ns.class_(
    "Sen0395UartPresenceSwitch", DfrobotSen0395Switch
)
Sen0395StartAfterBootSwitch = dfrobot_sen0395_ns.class_(
    "Sen0395StartAfterBootSwitch", DfrobotSen0395Switch
)


def _switch_schema(class_: MockObjClass) -> cv.Schema:
    return (
        switch.switch_schema(
            class_,
            entity_category=ENTITY_CATEGORY_CONFIG,
        )
        .extend(
            {
                cv.GenerateID(CONF_DFROBOT_SEN0395_ID): cv.use_id(
                    DfrobotSen0395Component
                ),
            }
        )
        .extend(cv.COMPONENT_SCHEMA)
    )


CONFIG_SCHEMA = cv.typed_schema(
    {
        "sensor_active": _switch_schema(Sen0395PowerSwitch),
        "turn_on_led": _switch_schema(Sen0395LedSwitch),
        "presence_via_uart": _switch_schema(Sen0395UartPresenceSwitch),
        "start_after_boot": _switch_schema(Sen0395StartAfterBootSwitch),
    }
)


async def to_code(config):
    parent = await cg.get_variable(config[CONF_DFROBOT_SEN0395_ID])
    var = await switch.new_switch(config)
    await cg.register_component(var, config)
    await cg.register_parented(var, parent)
    cg.add(getattr(parent, f"set_{config[CONF_TYPE]}_switch")(var))



from esphome.const import CONF_IGNORE_STRAPPING_WARNING, CONF_NUMBER

from .ili import GC9A01A, ILI9341, ILI9342, ST7789V

ILI9341.extend(
    # ESP32-2432S028 CYD board with Micro USB, has ILI9341 controller
    "ESP32-2432S028",
    data_rate="40MHz",
    cs_pin={"number": 15, "ignore_strapping_warning": True},
    dc_pin={"number": 2, "ignore_strapping_warning": True},
)

ST7789V.extend(
    # ESP32-2432S028 CYD board with USB C + Micro USB, has ST7789V controller
    "ESP32-2432S028-7789",
    data_rate="40MHz",
    cs_pin={"number": 15, "ignore_strapping_warning": True},
    dc_pin={"number": 2, "ignore_strapping_warning": True},
)

# fmt: off

ILI9342.extend(
    # ESP32-2432S028 CYD board with USB C + Micro USB, has ILI9342 controller
    "ESP32-2432S028-9342",
    data_rate="40MHz",
    cs_pin={"number": 15, "ignore_strapping_warning": True},
    dc_pin={"number": 2, "ignore_strapping_warning": True},
    initsequence=(
        (0xCB, 0x39, 0x2C, 0x00, 0x34, 0x02),  # Power Control A
        (0xCF, 0x00, 0xC1, 0x30),  # Power Control B
        (0xE8, 0x85, 0x00, 0x78),  # Driver timing control A
        (0xEA, 0x00, 0x00),  # Driver timing control B
        (0xED, 0x64, 0x03, 0x12, 0x81),  # Power on sequence control
        (0xF7, 0x20),  # Pump ratio control
        (0xC0, 0x23),  # Power Control 1
        (0xC1, 0x10),  # Power Control 2
        (0xC5, 0x3E, 0x28),  # VCOM Control 1
        (0xC7, 0x86),  # VCOM Control 2
        (0xB1, 0x00, 0x1B),  # Frame Rate Control
        (0xB6, 0x0A, 0xA2, 0x27, 0x00),  # Display Function Control
        (0xF2, 0x00),  # Enable 3G
        (0x26, 0x01),  # Gamma Set
        (0xE0, 0x00, 0x0C, 0x11, 0x04, 0x11, 0x08, 0x37, 0x89, 0x4C, 0x06, 0x0C, 0x0A, 0x2E, 0x34, 0x0F),  # Positive Gamma Correction
        (0xE1, 0x00, 0x0B, 0x11, 0x05, 0x13, 0x09, 0x33, 0x67, 0x48, 0x07, 0x0E, 0x0B, 0x23, 0x33, 0x0F),  # Negative Gamma Correction
    )
)

GC9A01A.extend(
    "ESP32-2424S012",
    invert_colors=True,
    cs_pin=10,
    dc_pin={CONF_NUMBER: 2, CONF_IGNORE_STRAPPING_WARNING: True},
)
